import abc
import time

import attr
import pyperf
import six

from ._to_bool import to_bool


var = attr.ib


def var_bool(*args, **kwargs):
    return attr.ib(*args, **kwargs, converter=to_bool)


def _register(scenario_cls):
    """Registers a scenario for benchmarking."""
    # This extends pyperf's runner by registering arguments for the scenario config
    def add_cmdline_args(cmd, args):
        for field in attr.fields(scenario_cls):
            if hasattr(args, field.name):
                cmd.extend(("--{}".format(field.name), str(getattr(args, field.name))))

    runner = pyperf.Runner(add_cmdline_args=add_cmdline_args)
    cmd = runner.argparser

    for field in attr.fields(scenario_cls):
        cmd.add_argument("--{}".format(field.name), type=field.type)

    parsed_args = runner.parse_args()

    config_dict = {
        field.name: getattr(parsed_args, field.name)
        for field in attr.fields(scenario_cls)
        if hasattr(parsed_args, field.name)
    }
    scenario = scenario_cls(**config_dict)

    runner.bench_time_func(scenario.scenario_name, scenario._pyperf)


class ScenarioMeta(abc.ABCMeta):
    def __init__(cls, name, bases, _dict):
        super(ScenarioMeta, cls).__init__(name, bases, _dict)

        # Make sure every sub-class is wrapped by `attr.s`
        cls = attr.s()(cls)

        # Do not register the base Scenario class
        # DEV: We cannot compare `cls` to `Scenario` since it doesn't exist yet
        if cls.__module__ != __name__:
            _register(cls)


class Scenario(six.with_metaclass(ScenarioMeta)):
    """The base class for specifying a benchmark."""

    name = attr.ib(type=str)

    @property
    def scenario_name(self):
        return "{}-{}".format(self.__class__.__name__.lower(), self.name)

    @abc.abstractmethod
    def run(self):
        """Returns a context manager that yields a function to be run for performance testing."""
        pass

    def _pyperf(self, loops):
        rungen = self.run()
        run = next(rungen)
        t0 = time.perf_counter()
        run(loops)
        dt = time.perf_counter() - t0
        try:
            # perform any teardown
            next(rungen)
        except StopIteration:
            pass
        finally:
            return dt



import os

import pymysql

from ddtrace import Pin
from ddtrace import config
from ddtrace.contrib.dbapi import TracedConnection
from ddtrace.vendor import wrapt

from ...ext import db
from ...ext import net
from ...internal.utils.formats import asbool


config._add(
    "pymysql",
    dict(
        # TODO[v1.0] this should be "mysql"
        _default_service="pymysql",
        _dbapi_span_name_prefix="pymysql",
        trace_fetch_methods=asbool(os.getenv("DD_PYMYSQL_TRACE_FETCH_METHODS", default=False)),
    ),
)

CONN_ATTR_BY_TAG = {
    net.TARGET_HOST: "host",
    net.TARGET_PORT: "port",
    db.USER: "user",
    db.NAME: "db",
}


def patch():
    wrapt.wrap_function_wrapper("pymysql", "connect", _connect)


def unpatch():
    if isinstance(pymysql.connect, wrapt.ObjectProxy):
        pymysql.connect = pymysql.connect.__wrapped__


def _connect(func, instance, args, kwargs):
    conn = func(*args, **kwargs)
    return patch_conn(conn)


def patch_conn(conn):
    tags = {t: getattr(conn, a, "") for t, a in CONN_ATTR_BY_TAG.items()}
    pin = Pin(tags=tags)

    # grab the metadata from the conn
    wrapped = TracedConnection(conn, pin=pin, cfg=config.pymysql)
    pin.onto(wrapped)
    return wrapped



"""
The httpx__ integration traces all HTTP requests made with the ``httpx``
library.

Enabling
~~~~~~~~

The ``httpx`` integration is enabled automatically when using
:ref:`ddtrace-run<ddtracerun>` or :func:`patch_all()<ddtrace.patch_all>`.

Alternatively, use :func:`patch()<ddtrace.patch>` to manually enable the integration::

    from ddtrace import patch
    patch(httpx=True)

    # use httpx like usual


Global Configuration
~~~~~~~~~~~~~~~~~~~~

.. py:data:: ddtrace.config.httpx['service']

   The default service name for ``httpx`` requests.
   By default the ``httpx`` integration will not define a service name and inherit
   its service name from its parent span.

   If you are making calls to uninstrumented third party applications you can
   set this setting, use the ``ddtrace.config.httpx['split_by_domain']`` setting,
   or use a ``Pin`` to override an individual connection's settings (see example
   below for ``Pin`` usage).

   This option can also be set with the ``DD_HTTPX_SERVICE`` environment
   variable.

   Default: ``None``


.. py:data:: ddtrace.config.httpx['distributed_tracing']

   Whether or not to inject distributed tracing headers into requests.

   Default: ``True``


.. py:data:: ddtrace.config.httpx['split_by_domain']

   Whether or not to use the domain name of requests as the service name. This
   setting can be overridden with session overrides (described in the Instance
   Configuration section).

   This setting takes precedence over ``ddtrace.config.httpx['service']``

   Default: ``False``


Instance Configuration
~~~~~~~~~~~~~~~~~~~~~~

To configure particular ``httpx`` client instances use the :class:`Pin <ddtrace.Pin>` API::

    import httpx
    from ddtrace import Pin

    client = httpx.Client()
    # Override service name for this instance
    Pin.override(client, service="custom-http-service")

    async_client = httpx.AsyncClient(
    # Override service name for this instance
    Pin.override(async_client, service="custom-async-http-service")


:ref:`Headers tracing <http-headers-tracing>` is supported for this integration.

:ref:`HTTP Tagging <http-tagging>` is supported for this integration.

.. __: https://www.python-httpx.org/
"""
from ...internal.utils.importlib import require_modules


required_modules = ["httpx"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from .patch import patch
        from .patch import unpatch

        __all__ = [
            "patch",
            "unpatch",
        ]



import os
import threading
from time import sleep

import pytest

from ddtrace.internal import periodic
from ddtrace.internal import service


if os.getenv("DD_PROFILE_TEST_GEVENT", False):
    import gevent

    class Event(object):
        """
        We can't use gevent Events here[0], nor can we use native threading
        events (because gevent is not multi-threaded).

        So for gevent, since it's not multi-threaded and will not run greenlets
        in parallel (for our usage here, anyway) we can write a dummy Event
        class which just does a simple busy wait on a shared variable.

        [0] https://github.com/gevent/gevent/issues/891
        """

        state = False

        def wait(self):
            while not self.state:
                gevent.sleep(0.001)

        def set(self):
            self.state = True


else:
    Event = threading.Event


def test_periodic():
    x = {"OK": False}

    thread_started = Event()
    thread_continue = Event()

    def _run_periodic():
        thread_started.set()
        x["OK"] = True
        thread_continue.wait()

    def _on_shutdown():
        x["DOWN"] = True

    t = periodic.PeriodicRealThreadClass()(0.001, _run_periodic, on_shutdown=_on_shutdown)
    t.start()
    thread_started.wait()
    thread_continue.set()
    assert t.is_alive()
    t.stop()
    t.join()
    assert not t.is_alive()
    assert x["OK"]
    assert x["DOWN"]
    if hasattr(threading, "get_native_id"):
        assert t.native_id is not None


def test_periodic_double_start():
    def _run_periodic():
        pass

    t = periodic.PeriodicRealThreadClass()(0.1, _run_periodic)
    t.start()
    with pytest.raises(RuntimeError):
        t.start()


def test_periodic_error():
    x = {"OK": False}

    thread_started = Event()
    thread_continue = Event()

    def _run_periodic():
        thread_started.set()
        thread_continue.wait()
        raise ValueError

    def _on_shutdown():
        x["DOWN"] = True

    t = periodic.PeriodicRealThreadClass()(0.001, _run_periodic, on_shutdown=_on_shutdown)
    t.start()
    thread_started.wait()
    thread_continue.set()
    t.stop()
    t.join()
    assert "DOWN" not in x


def test_gevent_class():
    if os.getenv("DD_PROFILE_TEST_GEVENT", False):
        assert isinstance(periodic.PeriodicRealThreadClass()(1, sum), periodic._GeventPeriodicThread)
    else:
        assert isinstance(periodic.PeriodicRealThreadClass()(1, sum), periodic.PeriodicThread)


def test_periodic_service_start_stop():
    t = periodic.PeriodicService(1)
    t.start()
    with pytest.raises(service.ServiceStatusError):
        t.start()
    t.stop()
    t.join()
    with pytest.raises(service.ServiceStatusError):
        t.stop()
    with pytest.raises(service.ServiceStatusError):
        t.stop()
    t.join()
    t.join()


def test_periodic_join_stop_no_start():
    t = periodic.PeriodicService(1)
    t.join()
    with pytest.raises(service.ServiceStatusError):
        t.stop()
    t.join()
    t = periodic.PeriodicService(1)
    with pytest.raises(service.ServiceStatusError):
        t.stop()
    t.join()
    with pytest.raises(service.ServiceStatusError):
        t.stop()


def test_is_alive_before_start():
    def x():
        pass

    t = periodic.PeriodicRealThreadClass()(1, x)
    assert not t.is_alive()


def test_awakeable_periodic_service():
    queue = []

    class AwakeMe(periodic.AwakeablePeriodicService):
        def periodic(self):
            queue.append(len(queue))

    interval = 1

    awake_me = AwakeMe(interval)

    awake_me.start()

    # Manually awake the service
    n = 10
    for _ in range(10):
        awake_me.awake()

    # Sleep long enough to also trigger the periodic function with the timeout
    sleep(1.1 * interval)

    awake_me.stop()

    assert queue == list(range(n + 2))



import contextlib
import logging
from threading import Event
import unittest

from cassandra.cluster import Cluster
from cassandra.cluster import ResultSet
from cassandra.query import BatchStatement
from cassandra.query import SimpleStatement

from ddtrace import Pin
from ddtrace import config
from ddtrace.constants import ANALYTICS_SAMPLE_RATE_KEY
from ddtrace.constants import ERROR_MSG
from ddtrace.constants import ERROR_TYPE
from ddtrace.contrib.cassandra.patch import patch
from ddtrace.contrib.cassandra.patch import unpatch
from ddtrace.contrib.cassandra.session import SERVICE
from ddtrace.ext import cassandra as cassx
from ddtrace.ext import net
from tests.contrib.config import CASSANDRA_CONFIG
from tests.opentracer.utils import init_tracer
from tests.utils import DummyTracer
from tests.utils import TracerTestCase
from tests.utils import assert_is_measured


# Oftentimes our tests fails because Cassandra connection timeouts during keyspace drop. Slowness in keyspace drop
# is known and is due to 'auto_snapshot' configuration. In our test env we should disable it, but the official cassandra
# image that we are using only allows us to configure a few configs:
# https://github.com/docker-library/cassandra/blob/4474c6c5cc2a81ee57c5615aae00555fca7e26a6/3.11/docker-entrypoint.sh#L51
# So for now we just increase the timeout, if this is not enough we may want to extend the official image with our own
# custom image.
CONNECTION_TIMEOUT_SECS = 20  # override the default value of 5

logging.getLogger("cassandra").setLevel(logging.INFO)


def setUpModule():
    # skip all the modules if the Cluster is not available
    if not Cluster:
        raise unittest.SkipTest("cassandra.cluster.Cluster is not available.")

    # create the KEYSPACE for this test module
    cluster = Cluster(port=CASSANDRA_CONFIG["port"], connect_timeout=CONNECTION_TIMEOUT_SECS)
    session = cluster.connect()
    session.execute("DROP KEYSPACE IF EXISTS test", timeout=10)
    session.execute(
        "CREATE KEYSPACE if not exists test WITH REPLICATION = { 'class' : 'SimpleStrategy', 'replication_factor': 1};"
    )
    session.execute("CREATE TABLE if not exists test.person (name text PRIMARY KEY, age int, description text)")
    session.execute("CREATE TABLE if not exists test.person_write (name text PRIMARY KEY, age int, description text)")
    session.execute("INSERT INTO test.person (name, age, description) VALUES ('Cassandra', 100, 'A cruel mistress')")
    session.execute(
        "INSERT INTO test.person (name, age, description) VALUES ('Athena', 100, 'Whose shield is thunder')"
    )
    session.execute("INSERT INTO test.person (name, age, description) VALUES ('Calypso', 100, 'Softly-braided nymph')")


def tearDownModule():
    # destroy the KEYSPACE
    cluster = Cluster(port=CASSANDRA_CONFIG["port"], connect_timeout=CONNECTION_TIMEOUT_SECS)
    session = cluster.connect()
    session.execute("DROP TABLE IF EXISTS test.person")
    session.execute("DROP TABLE IF EXISTS test.person_write")
    session.execute("DROP KEYSPACE IF EXISTS test", timeout=10)


class CassandraBase(object):
    """
    Needs a running Cassandra
    """

    TEST_QUERY = "SELECT * from test.person WHERE name = 'Cassandra'"
    TEST_QUERY_PAGINATED = "SELECT * from test.person"
    TEST_KEYSPACE = "test"
    TEST_PORT = CASSANDRA_CONFIG["port"]
    TEST_SERVICE = "test-cassandra"

    def _traced_session(self):
        # implement me
        pass

    @contextlib.contextmanager
    def override_config(self, integration, values):
        """
        Temporarily override an integration configuration value
        >>> with self.override_config('flask', dict(service_name='test-service')):
        ... # Your test
        """
        options = getattr(config, integration)

        original = dict((key, options.get(key)) for key in values.keys())

        options.update(values)
        try:
            yield
        finally:
            options.update(original)

    def setUp(self):
        self.cluster = Cluster(port=CASSANDRA_CONFIG["port"])
        self.session = self.cluster.connect()

    def _assert_result_correct(self, result):
        assert len(result.current_rows) == 1
        for r in result:
            assert r.name == "Cassandra"
            assert r.age == 100
            assert r.description == "A cruel mistress"

    def _test_query_base(self, execute_fn):
        session, tracer = self._traced_session()

        result = execute_fn(session, self.TEST_QUERY)
        self._assert_result_correct(result)

        spans = tracer.pop()
        assert spans, spans

        # another for the actual query
        assert len(spans) == 1

        query = spans[0]

        assert_is_measured(query)
        assert query.service == self.TEST_SERVICE
        assert query.resource == self.TEST_QUERY
        assert query.span_type == "cassandra"

        assert query.get_tag(cassx.KEYSPACE) == self.TEST_KEYSPACE
        assert query.get_metric(net.TARGET_PORT) == self.TEST_PORT
        assert query.get_metric(cassx.ROW_COUNT) == 1
        assert query.get_tag(cassx.PAGE_NUMBER) is None
        assert query.get_tag(cassx.PAGINATED) == "False"
        assert query.get_tag(net.TARGET_HOST) == "127.0.0.1"

        # confirm no analytics sample rate set by default
        assert query.get_metric(ANALYTICS_SAMPLE_RATE_KEY) is None

    def test_query(self):
        def execute_fn(session, query):
            return session.execute(query)

        self._test_query_base(execute_fn)

    def test_query_analytics_with_rate(self):
        with self.override_config("cassandra", dict(analytics_enabled=True, analytics_sample_rate=0.5)):
            session, tracer = self._traced_session()
            session.execute(self.TEST_QUERY)

            spans = tracer.pop()
            assert spans, spans
            # another for the actual query
            assert len(spans) == 1
            query = spans[0]
            # confirm no analytics sample rate set by default
            assert query.get_metric(ANALYTICS_SAMPLE_RATE_KEY) == 0.5

    def test_query_analytics_without_rate(self):
        with self.override_config("cassandra", dict(analytics_enabled=True)):
            session, tracer = self._traced_session()
            session.execute(self.TEST_QUERY)

            spans = tracer.pop()
            assert spans, spans
            # another for the actual query
            assert len(spans) == 1
            query = spans[0]
            # confirm no analytics sample rate set by default
            assert query.get_metric(ANALYTICS_SAMPLE_RATE_KEY) == 1.0

    def test_query_ot(self):
        """Ensure that cassandra works with the opentracer."""

        def execute_fn(session, query):
            return session.execute(query)

        session, tracer = self._traced_session()
        ot_tracer = init_tracer("cass_svc", tracer)

        with ot_tracer.start_active_span("cass_op"):
            result = execute_fn(session, self.TEST_QUERY)
            self._assert_result_correct(result)

        spans = tracer.pop()
        assert spans, spans

        # another for the actual query
        assert len(spans) == 2
        ot_span, dd_span = spans

        # confirm parenting
        assert ot_span.parent_id is None
        assert dd_span.parent_id == ot_span.span_id

        assert ot_span.name == "cass_op"
        assert ot_span.service == "cass_svc"

        assert dd_span.service == self.TEST_SERVICE
        assert dd_span.resource == self.TEST_QUERY
        assert dd_span.span_type == "cassandra"

        assert dd_span.get_tag(cassx.KEYSPACE) == self.TEST_KEYSPACE
        assert dd_span.get_metric(net.TARGET_PORT) == self.TEST_PORT
        assert dd_span.get_metric(cassx.ROW_COUNT) == 1
        assert dd_span.get_tag(cassx.PAGE_NUMBER) is None
        assert dd_span.get_tag(cassx.PAGINATED) == "False"
        assert dd_span.get_tag(net.TARGET_HOST) == "127.0.0.1"

    def test_query_async(self):
        def execute_fn(session, query):
            event = Event()
            result = []
            future = session.execute_async(query)

            def callback(results):
                result.append(ResultSet(future, results))
                event.set()

            future.add_callback(callback)
            event.wait()
            return result[0]

        self._test_query_base(execute_fn)

    def test_query_async_clearing_callbacks(self):
        def execute_fn(session, query):
            future = session.execute_async(query)
            future.clear_callbacks()
            return future.result()

        self._test_query_base(execute_fn)

    def test_span_is_removed_from_future(self):
        session, tracer = self._traced_session()
        future = session.execute_async(self.TEST_QUERY)
        future.result()
        span = getattr(future, "_ddtrace_current_span", None)
        assert span is None

    def test_paginated_query(self):
        session, tracer = self._traced_session()

        statement = SimpleStatement(self.TEST_QUERY_PAGINATED, fetch_size=1)
        result = session.execute(statement)
        # iterate over all pages
        results = list(result)
        assert len(results) == 3

        spans = tracer.pop()
        assert spans, spans

        # There are 4 spans for 3 results since the driver makes a request with
        # no result to check that it has reached the last page
        assert len(spans) == 4

        for i in range(4):
            query = spans[i]
            assert query.service == self.TEST_SERVICE
            assert query.resource == self.TEST_QUERY_PAGINATED
            assert query.span_type == "cassandra"

            assert query.get_tag(cassx.KEYSPACE) == self.TEST_KEYSPACE
            assert query.get_metric(net.TARGET_PORT) == self.TEST_PORT
            if i == 3:
                assert query.get_metric(cassx.ROW_COUNT) == 0
            else:
                assert query.get_metric(cassx.ROW_COUNT) == 1
            assert query.get_tag(net.TARGET_HOST) == "127.0.0.1"
            assert query.get_tag(cassx.PAGINATED) == "True"
            assert query.get_metric(cassx.PAGE_NUMBER) == i + 1

    def test_trace_with_service(self):
        session, tracer = self._traced_session()

        session.execute(self.TEST_QUERY)
        spans = tracer.pop()
        assert spans
        assert len(spans) == 1
        query = spans[0]
        assert query.service == self.TEST_SERVICE

    def test_trace_error(self):
        session, tracer = self._traced_session()

        try:
            session.execute("select * from test.i_dont_exist limit 1")
        except Exception:
            pass
        else:
            assert 0

        spans = tracer.pop()
        assert spans
        query = spans[0]
        assert query.error == 1
        for k in (ERROR_MSG, ERROR_TYPE):
            assert query.get_tag(k)

    def test_bound_statement(self):
        session, tracer = self._traced_session()

        query = "INSERT INTO test.person_write (name, age, description) VALUES (?, ?, ?)"
        prepared = session.prepare(query)
        session.execute(prepared, ("matt", 34, "can"))

        prepared = session.prepare(query)
        bound_stmt = prepared.bind(("leo", 16, "fr"))
        session.execute(bound_stmt)

        spans = tracer.pop()
        assert len(spans) == 2
        for s in spans:
            assert s.resource == query

    def test_batch_statement(self):
        session, tracer = self._traced_session()

        batch = BatchStatement()
        batch.add(
            SimpleStatement("INSERT INTO test.person_write (name, age, description) VALUES (%s, %s, %s)"),
            ("Joe", 1, "a"),
        )
        batch.add(
            SimpleStatement("INSERT INTO test.person_write (name, age, description) VALUES (%s, %s, %s)"),
            ("Jane", 2, "b"),
        )
        session.execute(batch)

        spans = tracer.pop()
        assert len(spans) == 1
        s = spans[0]
        assert s.resource == "BatchStatement"
        assert s.get_metric("cassandra.batch_size") == 2
        assert "test.person" in s.get_tag("cassandra.query")

    def test_batched_bound_statement(self):
        session, tracer = self._traced_session()

        batch = BatchStatement()

        prepared_statement = session.prepare("INSERT INTO test.person_write (name, age, description) VALUES (?, ?, ?)")
        batch.add(prepared_statement.bind(("matt", 34, "can")))
        session.execute(batch)

        spans = tracer.pop()
        assert len(spans) == 1
        s = spans[0]
        assert s.resource == "BatchStatement"
        assert s.get_tag("cassandra.query") == ""


class TestCassPatchDefault(unittest.TestCase, CassandraBase):
    """Test Cassandra instrumentation with patching and default configuration"""

    TEST_SERVICE = SERVICE

    def tearDown(self):
        unpatch()

    def setUp(self):
        CassandraBase.setUp(self)
        patch()

    def _traced_session(self):
        tracer = DummyTracer()
        Pin.get_from(self.cluster).clone(tracer=tracer).onto(self.cluster)
        return self.cluster.connect(self.TEST_KEYSPACE), tracer


class TestCassPatchAll(TestCassPatchDefault):
    """Test Cassandra instrumentation with patching and custom service on all clusters"""

    TEST_SERVICE = "test-cassandra-patch-all"

    def tearDown(self):
        unpatch()

    def setUp(self):
        CassandraBase.setUp(self)
        patch()

    def _traced_session(self):
        tracer = DummyTracer()
        # pin the global Cluster to test if they will conflict
        Pin(service=self.TEST_SERVICE, tracer=tracer).onto(Cluster)
        self.cluster = Cluster(port=CASSANDRA_CONFIG["port"])

        return self.cluster.connect(self.TEST_KEYSPACE), tracer


class TestCassPatchOne(TestCassPatchDefault):
    """Test Cassandra instrumentation with patching and custom service on one cluster"""

    TEST_SERVICE = "test-cassandra-patch-one"

    def tearDown(self):
        unpatch()

    def setUp(self):
        CassandraBase.setUp(self)
        patch()

    def _traced_session(self):
        tracer = DummyTracer()
        # pin the global Cluster to test if they will conflict
        Pin(service="not-%s" % self.TEST_SERVICE).onto(Cluster)
        self.cluster = Cluster(port=CASSANDRA_CONFIG["port"])

        Pin(service=self.TEST_SERVICE, tracer=tracer).onto(self.cluster)
        return self.cluster.connect(self.TEST_KEYSPACE), tracer

    def test_patch_unpatch(self):
        # Test patch idempotence
        patch()
        patch()

        tracer = DummyTracer()
        Pin.get_from(Cluster).clone(tracer=tracer).onto(Cluster)

        session = Cluster(port=CASSANDRA_CONFIG["port"]).connect(self.TEST_KEYSPACE)
        session.execute(self.TEST_QUERY)

        spans = tracer.pop()
        assert spans, spans
        assert len(spans) == 1

        # Test unpatch
        unpatch()

        session = Cluster(port=CASSANDRA_CONFIG["port"]).connect(self.TEST_KEYSPACE)
        session.execute(self.TEST_QUERY)

        spans = tracer.pop()
        assert not spans, spans

        # Test patch again
        patch()
        Pin.get_from(Cluster).clone(tracer=tracer).onto(Cluster)

        session = Cluster(port=CASSANDRA_CONFIG["port"]).connect(self.TEST_KEYSPACE)
        session.execute(self.TEST_QUERY)

        spans = tracer.pop()
        assert spans, spans


class TestCassandraConfig(TracerTestCase):
    """
    Test various configurations of the Cassandra integration.
    """

    TEST_QUERY = "SELECT * from test.person WHERE name = 'Cassandra'"
    TEST_KEYSPACE = "test"

    def setUp(self):
        super(TestCassandraConfig, self).setUp()
        patch()
        self.tracer = DummyTracer()
        self.cluster = Cluster(port=CASSANDRA_CONFIG["port"])
        Pin.get_from(self.cluster).clone(tracer=self.tracer).onto(self.cluster)
        self.session = self.cluster.connect(self.TEST_KEYSPACE)

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_SERVICE="mysvc"))
    def test_user_specified_service(self):
        """
        When a user specifies a service for the app
            The cassandra integration should not use it.
        """
        # Ensure that the service name was configured
        from ddtrace import config

        assert config.service == "mysvc"

        self.session.execute(self.TEST_QUERY)
        spans = self.pop_spans()
        assert spans
        assert len(spans) == 1
        query = spans[0]
        assert query.service != "mysvc"



import abc

import six


class Propagator(six.with_metaclass(abc.ABCMeta)):
    @staticmethod
    @abc.abstractmethod
    def inject(span_context, carrier):
        pass

    @staticmethod
    @abc.abstractmethod
    def extract(carrier):
        pass



import pytest

from ddtrace.internal.remoteconfig import RemoteConfig


def test_remote_config_register_auto_enable():
    assert RemoteConfig._worker is None

    RemoteConfig.register("LIVE_DEBUGGER", lambda m, c: None)

    assert RemoteConfig._worker is not None

    RemoteConfig.disable()

    assert RemoteConfig._worker is None


@pytest.mark.subprocess
def test_remote_config_forksafe():
    import os

    from ddtrace.internal.remoteconfig import RemoteConfig

    RemoteConfig.enable()

    parent_worker = RemoteConfig._worker
    assert parent_worker is not None

    if os.fork() == 0:
        assert RemoteConfig._worker is not None
        assert RemoteConfig._worker is not parent_worker
        exit(0)



"""
Tagset eBNF::

    tagset = tag, { ",", tag };
    tag = key, "=", value;
    key = { ? ASCII 32-126 ? - equal or comma or space };
    value = { ? ASCII 32-126 ? - comma };
    equal or comma = "=" | ",";
    space = " ";
"""

class TagsetEncodeError(ValueError):
    pass


class TagsetMaxSizeEncodeError(TagsetEncodeError):
    """Exception used when the encoded values exceed the max allowed size

    This exception will contain the:

    - original/in-tact values that are being encoded
    - the max size configured
    - the valid/encoded tagset up until the max size would be exceeded

    Example::

        try:
            tagset = encode_tagset_values({"a": "1", "b": 2, "c": "3"}, max_size=6)
        except TagsetMaxSizeEncodeError as e:
            assert e.values = {"a": "1", "b": "2", "c": "3"}
            assert e.max_size = 6
            assert e.current_results = "a=1,b=2"
    """
    def __init__(self, values, max_size, current_results):
        # type: (Dict[str, str], int, str) -> None
        self.values = values
        self.max_size = max_size
        self.current_results = current_results

        msg = "Tagset cannot encode {!r} would exceed max size of {!r}".format(values, max_size)
        super(TagsetMaxSizeEncodeError, self).__init__(msg)


class TagsetDecodeError(ValueError):
    pass


class TagsetMaxSizeDecodeError(TagsetDecodeError):
    """Exception used when tagset compatible string is greater than the max size

    This exception will contain the:

    - tagset string being decoded
    - the max size configured

    Example::

        try:
            tagset = decode_tagset_values("key=value", max_size=8)
        except TagsetMaxSizeDecodeError as e:
            assert e.tagset = "key=value"
            assert e.max_size = 8
    """
    def __init__(self, tagset, max_size):
        # type: (Dict[str, str], int, str) -> None
        self.tagset = tagset
        self.max_size = max_size

        msg = "Tagset cannot decode {!r} which exceeds max size of {!r} with length {!r}".format(tagset, max_size, len(tagset))
        super(TagsetMaxSizeDecodeError, self).__init__(msg)


cdef inline int is_equal(int c):
    # '='
    return c == 61


cdef inline int is_comma(int c):
    # ','
    return c == 44


cdef inline int is_valid_key_char(int c):
    # string.printable - " ,="
    # 32 = " "
    # 44 = ",""
    # 61 = "="
    return (33 <= c <= 43) or (45 <= c <= 60) or (62 <= c <= 126)


# Same as is_valid_key_char except spaces are allowed
cdef inline int is_valid_value_char(int c):
    # string.printable - ","
    # 44 = ",""
    return c == 32 or is_equal(c) or is_valid_key_char(c)


cpdef dict decode_tagset_string(str tagset, int max_size=512):
    # type: (str, int) -> Dict[str, str]
    """Parse a tagset compatible string into a dictionary of tag key/values

    Examples::

        >>> decode_tagset_string("key=value")
        {"key": "value"}
        >>> decode_tagset_string("a=1,b=2,c=3")
        {"a": "1", "b": "2", "c": "3"}

    :param str tagset: String of encoded "key=value" pairs to decode into a dictionary
    :param int max_size: The max size allowed for length of the incoming string (default: 512)
    :rtype: dict
    :returns: a Dict[str,str] of decoded key/value pairs from the provided string
    :raises TagsetDecodeError: When the provided format is not valid
    """
    cdef dict res = {}
    cdef str c
    cdef int o
    cdef str cur_val = ""
    cdef str cur_key = ""
    cdef int is_parsing_key = 1

    # No tagset provided, short circuit the response
    if not tagset:
        return res

    # Raise an exception that the incoming tagset string exceeds the max size
    if len(tagset) > max_size:
        raise TagsetMaxSizeDecodeError(tagset, max_size)

    # DEV: Parse in a single pass of `tagset`
    #      `is_parsing_key` is used to know if we are on
    #      right or left side of an `=`
    for c in tagset:
        o = ord(c)
        if is_parsing_key:
            if is_equal(o):
                if not cur_key:
                    raise TagsetDecodeError("Empty keys are not allowed: {!r}".format(tagset))
                is_parsing_key = 0
                continue

            if not is_valid_key_char(o):
                raise TagsetDecodeError("Unexpected {!r} character for key {}: {!r}".format(c, cur_key, tagset))
            cur_key += c
        else:
            if is_comma(o):
                cur_val = cur_val.strip(" ")
                if not cur_val:
                    raise TagsetDecodeError("Empty values are not allowed: {!r}".format(tagset))
                res[cur_key] = cur_val
                cur_key, cur_val, is_parsing_key = "", "", 1
                continue

            if not is_valid_value_char(o):
                raise TagsetDecodeError("Unexpected character {!r} for value {}={}: {!r}".format(c, cur_key, cur_val, tagset))
            cur_val += c

    # Reached EOF, do we have a key/value pair we need to save?
    if cur_key:
        cur_val = cur_val.strip(" ")
        if cur_val:
            res[cur_key] = cur_val
        else:
            raise TagsetDecodeError("Expected value for key {!r} instead got EOF: {!r}".format(cur_key, tagset))

    return res

cdef bint _key_is_valid(str key):
    """Helper to ensure a key's characters are all valid"""
    if not key:
        return 0

    for c in key:
        if not is_valid_key_char(ord(c)):
            return 0
    return 1


cdef bint _value_is_valid(str value):
    """Helper to ensure a values's characters are all valid"""
    if not value:
        return 0

    for c in value:
        if not is_valid_value_char(ord(c)):
            return 0
    return 1


cpdef str encode_tagset_values(object values, int max_size=512):
    # type: (Dict[str, str], int) -> str
    """Convert a dictionary of tag key/values into a tagset compatible string

    Example::

        >>> encode_tagset_values({"key": "value"})
        "key=value"
        >>> encode_tagset_values({"a": "1", "b": "2", "c": "3"})
        "a=1,b=2,c=3"

    :param dict values: Dict[str,str] of key/value pairs to encode
    :param int max_size: The max size to allow the resulting string to be (default: 512)
    :rtype: str
    :returns: tagset encoded key/value pairs
    :raises TagsetMaxSizeEncodeError: Raised when we will exceed the provided max size
    :raises TagsetEncodeError: Raised when we encounter an exception character in a key or value
    """
    cdef str res = ""
    cdef str key
    cdef str value
    cdef int i

    for i, (key, value) in enumerate(values.items()):
        # Strip any leading/trailing spaces
        key = key.strip(" ")
        value = value.strip(" ")

        if not _key_is_valid(key):
            raise TagsetEncodeError("Key is not valid: {!r}".format(key))
        if not _value_is_valid(value):
            raise TagsetEncodeError("Value is not valid: {!r}".format(value))

        encoded = "{}={}".format(key, value)
        # Prefix every item except the first with `,` for separator
        if i > 0:
            encoded = "," + encoded

        # Raise an exception that we will exceed the max size
        # The exception has the value up until now if the caller
        # wants to use the partially encoded value
        if len(res) + len(encoded) > max_size:
            raise TagsetMaxSizeEncodeError(values, max_size, res)

        res += encoded
    return res



"""
async tracing utils

Note that this module should only be imported in Python 3.5+.
"""
from ddtrace import Pin
from ddtrace.internal.logger import get_logger


log = get_logger(__name__)


def with_traced_module(func):
    """Async version of trace_utils.with_traced_module.
    Usage::

        @with_traced_module
        async def my_traced_wrapper(django, pin, func, instance, args, kwargs):
            # Do tracing stuff
            pass

        def patch():
            import django
            wrap(django.somefunc, my_traced_wrapper(django))
    """

    def with_mod(mod):
        async def wrapper(wrapped, instance, args, kwargs):
            pin = Pin._find(instance, mod)
            if pin and not pin.enabled():
                return await wrapped(*args, **kwargs)
            elif not pin:
                log.debug("Pin not found for traced method %r", wrapped)
                return await wrapped(*args, **kwargs)
            return await func(mod, pin, wrapped, instance, args, kwargs)

        return wrapper

    return with_mod



# -*- encoding: utf-8 -*-
import binascii
import datetime
import gzip
import itertools
import json
import os
import platform
import typing

import attr
import six
from six.moves import http_client
import tenacity

import ddtrace
from ddtrace.internal import agent
from ddtrace.internal import runtime
from ddtrace.internal.runtime import container
from ddtrace.internal.utils import attr as attr_utils
from ddtrace.internal.utils.formats import parse_tags_str
from ddtrace.profiling import exporter
from ddtrace.profiling import recorder
from ddtrace.profiling.exporter import pprof


HOSTNAME = platform.node()
PYTHON_IMPLEMENTATION = platform.python_implementation().encode()
PYTHON_VERSION = platform.python_version().encode()


class UploadFailed(tenacity.RetryError, exporter.ExportError):
    """Upload failure."""

    def __str__(self):
        return str(self.last_attempt.exception())


@attr.s
class PprofHTTPExporter(pprof.PprofExporter):
    """PProf HTTP exporter."""

    # repeat this to please mypy
    enable_code_provenance = attr.ib(default=True, type=bool)

    endpoint = attr.ib(type=str, factory=agent.get_trace_url)
    api_key = attr.ib(default=None, type=typing.Optional[str])
    # Do not use the default agent timeout: it is too short, the agent is just a unbuffered proxy and the profiling
    # backend is not as fast as the tracer one.
    timeout = attr.ib(factory=attr_utils.from_env("DD_PROFILING_API_TIMEOUT", 10.0, float), type=float)
    service = attr.ib(default=None, type=typing.Optional[str])
    env = attr.ib(default=None, type=typing.Optional[str])
    version = attr.ib(default=None, type=typing.Optional[str])
    tags = attr.ib(factory=dict, type=typing.Dict[str, bytes])
    max_retry_delay = attr.ib(default=None)
    _container_info = attr.ib(factory=container.get_container_info, repr=False)
    _retry_upload = attr.ib(init=False, eq=False)
    endpoint_path = attr.ib(default="/profiling/v1/input")

    def __attrs_post_init__(self):
        if self.max_retry_delay is None:
            self.max_retry_delay = self.timeout * 3
        self._retry_upload = tenacity.Retrying(
            # Retry after 1s, 2s, 4s, 8s with some randomness
            wait=tenacity.wait_random_exponential(multiplier=0.5),
            stop=tenacity.stop_after_delay(self.max_retry_delay),
            retry_error_cls=UploadFailed,
            retry=tenacity.retry_if_exception_type((http_client.HTTPException, OSError, IOError)),
        )
        tags = {
            k: six.ensure_binary(v)
            for k, v in itertools.chain(
                parse_tags_str(os.environ.get("DD_TAGS")).items(),
                parse_tags_str(os.environ.get("DD_PROFILING_TAGS")).items(),
            )
        }
        tags.update({k: six.ensure_binary(v) for k, v in self.tags.items()})
        tags.update(
            {
                "host": HOSTNAME.encode("utf-8"),
                "language": b"python",
                "runtime": PYTHON_IMPLEMENTATION,
                "runtime_version": PYTHON_VERSION,
                "profiler_version": ddtrace.__version__.encode("ascii"),
            }
        )
        if self.version:
            tags["version"] = self.version.encode("utf-8")

        if self.env:
            tags["env"] = self.env.encode("utf-8")

        self.tags = tags

    @staticmethod
    def _encode_multipart_formdata(
        fields,  # type: typing.Dict[str, bytes]
        tags,  # type: typing.Dict[str, bytes]
        data,  # type: typing.Dict[bytes, bytes]
    ):
        # type: (...) -> typing.Tuple[bytes, bytes]
        boundary = binascii.hexlify(os.urandom(16))

        # The body that is generated is very sensitive and must perfectly match what the server expects.
        body = (
            b"".join(
                b"--%s\r\n"
                b'Content-Disposition: form-data; name="%s"\r\n'
                b"\r\n"
                b"%s\r\n" % (boundary, field.encode(), value)
                for field, value in fields.items()
                if field != "chunk-data"
            )
            + b"".join(
                b"--%s\r\n"
                b'Content-Disposition: form-data; name="tags[]"\r\n'
                b"\r\n"
                b"%s:%s\r\n" % (boundary, tag.encode(), value)
                for tag, value in tags.items()
            )
            + b"".join(
                (
                    b'--%s\r\nContent-Disposition: form-data; name="data[%s]"; filename="%s"\r\n'
                    % (boundary, field_name, field_name)
                )
                + b"Content-Type: application/octet-stream\r\n\r\n"
                + field_data
                + b"\r\n"
                for field_name, field_data in data.items()
            )
            + b"--%s--" % boundary
        )

        content_type = b"multipart/form-data; boundary=%s" % boundary

        return content_type, body

    def _get_tags(
        self, service  # type: str
    ):
        # type: (...) -> typing.Dict[str, bytes]
        tags = {
            "service": service.encode("utf-8"),
            "runtime-id": runtime.get_runtime_id().encode("ascii"),
        }

        tags.update(self.tags)

        return tags

    def export(
        self,
        events,  # type: recorder.EventsType
        start_time_ns,  # type: int
        end_time_ns,  # type: int
    ):
        # type: (...) -> typing.Tuple[pprof.pprof_ProfileType, typing.List[pprof.Package]]
        """Export events to an HTTP endpoint.

        :param events: The event dictionary from a `ddtrace.profiling.recorder.Recorder`.
        :param start_time_ns: The start time of recording.
        :param end_time_ns: The end time of recording.
        """
        if self.api_key:
            headers = {
                "DD-API-KEY": self.api_key.encode(),
            }
        else:
            headers = {}

        if self._container_info and self._container_info.container_id:
            headers["Datadog-Container-Id"] = self._container_info.container_id

        profile, libs = super(PprofHTTPExporter, self).export(events, start_time_ns, end_time_ns)
        pprof = six.BytesIO()
        with gzip.GzipFile(fileobj=pprof, mode="wb") as gz:
            gz.write(profile.SerializeToString())
        fields = {
            "version": b"3",
            "family": b"python",
            "runtime-id": runtime.get_runtime_id().encode("ascii"),
            "start": (
                datetime.datetime.utcfromtimestamp(start_time_ns / 1e9).replace(microsecond=0).isoformat() + "Z"
            ).encode(),
            "end": (
                datetime.datetime.utcfromtimestamp(end_time_ns / 1e9).replace(microsecond=0).isoformat() + "Z"
            ).encode(),
        }

        service = self.service or os.path.basename(profile.string_table[profile.mapping[0].filename])

        data = {b"auto.pprof": pprof.getvalue()}

        if self.enable_code_provenance:
            code_provenance = six.BytesIO()
            with gzip.GzipFile(fileobj=code_provenance, mode="wb") as gz:
                gz.write(
                    json.dumps(
                        {
                            "v1": libs,
                        }
                    ).encode("utf-8")
                )
            data[b"code-provenance.json"] = code_provenance.getvalue()

        content_type, body = self._encode_multipart_formdata(
            fields,
            tags=self._get_tags(service),
            data=data,
        )
        headers["Content-Type"] = content_type

        client = agent.get_connection(self.endpoint, self.timeout)
        self._upload(client, self.endpoint_path, body, headers)

        return profile, libs

    def _upload(self, client, path, body, headers):
        self._retry_upload(self._upload_once, client, path, body, headers)

    def _upload_once(self, client, path, body, headers):
        try:
            client.request("POST", path, body=body, headers=headers)
            response = client.getresponse()
            response.read()  # reading is mandatory
        finally:
            client.close()

        if 200 <= response.status < 300:
            return

        if 500 <= response.status < 600:
            raise tenacity.TryAgain

        if response.status == 400:
            raise exporter.ExportError("Server returned 400, check your API key")
        elif response.status == 404 and not self.api_key:
            raise exporter.ExportError(
                "Datadog Agent is not accepting profiles. "
                "Agent-based profiling deployments require Datadog Agent >= 7.20"
            )

        raise exporter.ExportError("HTTP Error %d" % response.status)



import django

from .. import views


if django.VERSION < (4, 0, 0):
    from django.conf.urls import url as handler
else:
    from django.urls import re_path as handler


urlpatterns = [
    handler(r"^$", views.index),
    handler(r"^simple/$", views.BasicView.as_view()),
]



from ddtrace.internal.codeowners import Codeowners


def test_invalid_codeowners(testdir):
    """Skip invalid lines and still match valid rules."""
    codeowners = """
    [invalid section
    * @default

    ^[invalid optional section
    bar.py @bars
    """
    codeowners_file = testdir.makefile("", CODEOWNERS=codeowners)

    c = Codeowners(path=codeowners_file.strpath)
    assert c.of("foo.py") == ["@default"]
    assert c.of("bar.py") == ["@bars"]



import os

from ddtrace.internal.packages import get_distributions


def test_get_distributions():
    """use pkg_resources to validate package names and versions returned by get_distributions()"""
    import pkg_resources

    pkg_resources_ws = {pkg.project_name for pkg in pkg_resources.working_set}

    importlib_pkgs = set()
    for pkg in get_distributions():
        assert pkg.name
        assert pkg.version
        assert os.path.exists(pkg.path)
        # The package name in typing_extensions-4.x.x.dist-info/METADATA is set to `typing_extensions`
        # this is inconsistent with the package name found in pkg_resources. The block below corrects this.
        # The correct package name is typing-extensions.
        # The issue exists in pkgutil-resolve-name package.
        if pkg.name == "typing_extensions" and "typing-extensions" in pkg_resources_ws:
            importlib_pkgs.add("typing-extensions")
        elif pkg.name == "pkgutil_resolve_name" and "pkgutil-resolve-name" in pkg_resources_ws:
            importlib_pkgs.add("pkgutil-resolve-name")
        else:
            importlib_pkgs.add(pkg.name)

    # assert that pkg_resources and importlib.metadata return the same packages
    assert pkg_resources_ws == importlib_pkgs



import aiopg

# project
from ddtrace import Pin
from ddtrace.contrib.aiopg.patch import patch
from ddtrace.contrib.aiopg.patch import unpatch
from tests.contrib.asyncio.utils import AsyncioTestCase
from tests.contrib.asyncio.utils import mark_asyncio
from tests.contrib.config import POSTGRES_CONFIG


TEST_PORT = str(POSTGRES_CONFIG["port"])


class AiopgTestCase(AsyncioTestCase):
    # default service
    TEST_SERVICE = "postgres"

    def setUp(self):
        super().setUp()
        self._conn = None
        patch()

    def tearDown(self):
        super().tearDown()
        if self._conn and not self._conn.closed:
            self._conn.close()

        unpatch()

    async def _get_conn_and_tracer(self):
        conn = self._conn = await aiopg.connect(**POSTGRES_CONFIG)
        Pin.get_from(conn).clone(tracer=self.tracer).onto(conn)

        return conn, self.tracer

    @mark_asyncio
    async def test_async_generator(self):
        conn, tracer = await self._get_conn_and_tracer()
        cursor = await conn.cursor()
        q = "select 'foobarblah'"
        await cursor.execute(q)
        rows = []
        async for row in cursor:
            rows.append(row)

        assert rows == [("foobarblah",)]
        spans = self.pop_spans()
        assert len(spans) == 1
        span = spans[0]
        assert span.name == "postgres.query"



from typing import TYPE_CHECKING

from ddtrace import Pin
from ddtrace import config
from ddtrace.vendor import wrapt

from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import db
from ...ext import net
from ...internal.logger import get_logger
from ...internal.utils import get_argument_value
from ..trace_utils import ext_service
from ..trace_utils import unwrap
from ..trace_utils import wrap
from ..trace_utils_async import with_traced_module


if TYPE_CHECKING:
    from types import ModuleType
    from typing import Dict
    from typing import Union

    import asyncpg
    from asyncpg.prepared_stmt import PreparedStatement


config._add(
    "asyncpg",
    dict(
        _default_service="postgres",
    ),
)


log = get_logger(__name__)


def _get_connection_tags(conn):
    # type: (asyncpg.Connection) -> Dict[str, str]
    addr = conn._addr
    params = conn._params
    host = port = ""
    if isinstance(addr, tuple) and len(addr) == 2:
        host, port = addr
    return {
        net.TARGET_HOST: host,
        net.TARGET_PORT: port,
        db.USER: params.user,
        db.NAME: params.database,
    }


class _TracedConnection(wrapt.ObjectProxy):
    def __init__(self, conn, pin):
        super(_TracedConnection, self).__init__(conn)
        conn_pin = pin.clone(tags=_get_connection_tags(conn))
        # Keep the pin on the protocol
        conn_pin.onto(self._protocol)

    def __setddpin__(self, pin):
        pin.onto(self._protocol)

    def __getddpin__(self):
        return Pin.get_from(self._protocol)


@with_traced_module
async def _traced_connect(asyncpg, pin, func, instance, args, kwargs):
    """Traced asyncpg.connect().

    connect() is instrumented and patched to return a connection proxy.
    """
    with pin.tracer.trace(
        "postgres.connect", span_type=SpanTypes.SQL, service=ext_service(pin, config.asyncpg)
    ) as span:
        # Need an ObjectProxy since Connection uses slots
        conn = _TracedConnection(await func(*args, **kwargs), pin)
        span.set_tags(_get_connection_tags(conn))
        return conn


async def _traced_query(pin, method, query, args, kwargs):
    with pin.tracer.trace(
        "postgres.query", resource=query, service=ext_service(pin, config.asyncpg), span_type=SpanTypes.SQL
    ) as span:
        span.set_tag(SPAN_MEASURED_KEY)
        span.set_tags(pin.tags)
        return await method(*args, **kwargs)


@with_traced_module
async def _traced_protocol_execute(asyncpg, pin, func, instance, args, kwargs):
    state = get_argument_value(args, kwargs, 0, "state")  # type: Union[str, PreparedStatement]
    query = state if isinstance(state, str) else state.query
    return await _traced_query(pin, func, query, args, kwargs)


def _patch(asyncpg):
    # type: (ModuleType) -> None
    wrap(asyncpg, "connect", _traced_connect(asyncpg))
    for method in ("execute", "bind_execute", "query", "bind_execute_many"):
        wrap(asyncpg.protocol, "Protocol.%s" % method, _traced_protocol_execute(asyncpg))


def patch():
    # type: () -> None
    import asyncpg

    if getattr(asyncpg, "_datadog_patch", False):
        return

    Pin().onto(asyncpg)
    _patch(asyncpg)

    setattr(asyncpg, "_datadog_patch", True)


def _unpatch(asyncpg):
    # type: (ModuleType) -> None
    unwrap(asyncpg, "connect")
    for method in ("execute", "bind_execute", "query", "bind_execute_many"):
        unwrap(asyncpg.protocol.Protocol, method)


def unpatch():
    # type: () -> None
    import asyncpg

    if not getattr(asyncpg, "_datadog_patch", False):
        return

    _unpatch(asyncpg)

    setattr(asyncpg, "_datadog_patch", False)



"""
pytest local plugin used to automatically make the following fixtures
available for all tests in this directory

https://docs.pytest.org/en/latest/writing_plugins.html#testing-plugins
"""
import pytest

from ddtrace.opentracer import Tracer
from ddtrace.opentracer import set_global_tracer
from tests.utils import DummyTracer
from tests.utils import TracerSpanContainer


@pytest.fixture()
def ot_tracer_factory():
    """Fixture which returns an opentracer ready to use for testing."""

    def make_ot_tracer(service_name="my_svc", config=None, scope_manager=None, context_provider=None):
        config = config or {}
        tracer = Tracer(service_name=service_name, config=config, scope_manager=scope_manager)

        # similar to how we test the ddtracer, use a dummy tracer
        dd_tracer = DummyTracer()
        if context_provider:
            dd_tracer.configure(context_provider=context_provider)

        # attach the dummy tracer to the opentracer
        tracer._dd_tracer = dd_tracer
        return tracer

    return make_ot_tracer


@pytest.fixture()
def ot_tracer(ot_tracer_factory):
    """Fixture for a default opentracer."""
    return ot_tracer_factory()


@pytest.fixture
def test_spans(ot_tracer):
    container = TracerSpanContainer(ot_tracer._dd_tracer)
    yield container
    container.reset()


@pytest.fixture()
def global_tracer(ot_tracer):
    """A function similar to one OpenTracing users would write to initialize
    their OpenTracing tracer.
    """
    set_global_tracer(ot_tracer)

    return ot_tracer


@pytest.fixture()
def dd_tracer(ot_tracer):
    return ot_tracer._dd_tracer



import os
import subprocess

import gevent
import pytest


@pytest.mark.skipif(
    not (gevent.version_info.major >= 1 and gevent.version_info.minor >= 3), reason="gevent 1.3 or later is required"
)
def test_gevent_warning(monkeypatch):
    subp = subprocess.Popen(
        ("python", os.path.join(os.path.dirname(__file__), "wrong_program_gevent.py")),
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
    )
    assert subp.wait() == 0
    assert subp.stdout.read() == b""
    assert b"RuntimeWarning: Loading ddtrace before using gevent monkey patching" in subp.stderr.read()


@pytest.mark.subprocess
def test_gevent_auto_patching():
    import ddtrace

    # Disable tracing sqlite3 as it is used by coverage
    ddtrace.patch_all(sqlite3=False)
    # Patch on import
    import gevent  # noqa

    from ddtrace.contrib.gevent import GeventContextProvider

    assert isinstance(ddtrace.tracer.context_provider, GeventContextProvider)


def test_gevent_ddtrace_run_auto_patching(ddtrace_run_python_code_in_subprocess):
    code = """
import gevent  # Patch on import
import ddtrace  # ddtrace-run, No need to call patch_all()
from ddtrace.contrib.gevent import GeventContextProvider


assert isinstance(ddtrace.tracer.context_provider, GeventContextProvider)
"""

    out, err, status, pid = ddtrace_run_python_code_in_subprocess(code)
    assert status == 0, err
    assert out == b""



from threading import RLock
from typing import Any
from typing import Callable
from typing import Optional
from typing import Type
from typing import TypeVar


miss = object()

T = TypeVar("T")
F = Callable[[T], Any]
M = Callable[[Any, T], Any]


class LFUCache(dict):
    """Simple LFU cache implementation.

    This cache is designed for memoizing functions with a single hashable
    argument. The eviction policy is LFU, i.e. the least frequently used values
    are evicted when the cache is full. The amortized cost of shrinking the
    cache when it grows beyond the requested size is O(log(size)).
    """

    def __init__(self, maxsize=256):
        # type: (int) -> None
        self.maxsize = maxsize
        self.lock = RLock()

    def get(self, key, f):  # type: ignore[override]
        # type: (T, F) -> Any
        """Get a value from the cache.

        If the value with the given key is not in the cache, the expensive
        function ``f`` is called on the key to generate it. The return value is
        then stored in the cache and returned to the caller.
        """
        if len(self) >= self.maxsize:
            for _, h in zip(range(self.maxsize >> 1), sorted(self, key=lambda h: self[h][1])):
                del self[h]

        _ = super(LFUCache, self).get(key, miss)
        if _ is not miss:
            value, count = _
            self[key] = (value, count + 1)
            return value

        with self.lock:
            _ = super(LFUCache, self).get(key, miss)
            if _ is not miss:
                value, count = _
                self[key] = (value, count + 1)
                return value

            value = f(key)

            self[key] = (value, 1)

            return value


def cached(maxsize=256):
    # type: (int) -> Callable[[F], F]
    """Decorator for memoizing functions of a single argument (LFU policy)."""

    def cached_wrapper(f):
        # type: (F) -> F
        cache = LFUCache(maxsize)

        def cached_f(key):
            # type: (T) -> Any
            return cache.get(key, f)

        cached_f.invalidate = cache.clear  # type: ignore[attr-defined]

        return cached_f

    return cached_wrapper


class CachedMethodDescriptor(object):
    def __init__(self, method, maxsize):
        # type: (M, int) -> None
        self._method = method
        self._maxsize = maxsize

    def __get__(self, obj, objtype=None):
        # type: (Any, Optional[Type]) -> F
        cached_method = cached(self._maxsize)(self._method.__get__(obj, objtype))
        setattr(obj, self._method.__name__, cached_method)
        return cached_method


def cachedmethod(maxsize=256):
    # type: (int) -> Callable[[M], CachedMethodDescriptor]
    """Decorator for memoizing methods of a single argument (LFU policy)."""

    def cached_wrapper(f):
        # type: (M) -> CachedMethodDescriptor
        return CachedMethodDescriptor(f, maxsize)

    return cached_wrapper



import ast
import contextlib
from itertools import product
import os
from os.path import split
from os.path import splitext
import sys
from tempfile import NamedTemporaryFile
import time

from _pytest.runner import CallInfo
from _pytest.runner import TestReport
from _pytest.runner import call_and_report
import pytest
from six import PY2

import ddtrace
from tests.utils import DummyTracer
from tests.utils import TracerSpanContainer
from tests.utils import call_program
from tests.utils import request_token
from tests.utils import snapshot_context as _snapshot_context


def pytest_configure(config):
    config.addinivalue_line(
        "markers", "snapshot(*args, **kwargs): mark test to run as a snapshot test which sends traces to the test agent"
    )


@pytest.fixture
def use_global_tracer():
    yield False


@pytest.fixture
def tracer(use_global_tracer):
    if use_global_tracer:
        return ddtrace.tracer
    else:
        return DummyTracer()


@pytest.fixture
def test_spans(tracer):
    container = TracerSpanContainer(tracer)
    yield container
    container.reset()


@pytest.fixture
def run_python_code_in_subprocess(tmpdir):
    def _run(code, **kwargs):
        pyfile = tmpdir.join("test.py")
        pyfile.write(code)
        return call_program(sys.executable, str(pyfile), **kwargs)

    yield _run


@pytest.fixture
def ddtrace_run_python_code_in_subprocess(tmpdir):
    def _run(code, **kwargs):
        pyfile = tmpdir.join("test.py")
        pyfile.write(code)
        return call_program("ddtrace-run", sys.executable, str(pyfile), **kwargs)

    yield _run


@pytest.fixture(autouse=True)
def snapshot(request):
    marks = [m for m in request.node.iter_markers(name="snapshot")]
    assert len(marks) < 2, "Multiple snapshot marks detected"
    if marks:
        snap = marks[0]
        token = snap.kwargs.get("token")
        if token:
            del snap.kwargs["token"]
        else:
            token = request_token(request).replace(" ", "_").replace(os.path.sep, "_")

        with _snapshot_context(token, *snap.args, **snap.kwargs) as snapshot:
            yield snapshot
    else:
        yield


@pytest.fixture
def snapshot_context(request):
    """
    Fixture to provide a context manager for executing code within a ``tests.utils.snapshot_context``
    with a default ``token`` based on the test function/``pytest`` request.

    def test_case(snapshot_context):
        with snapshot_context():
            # my code
    """
    token = request_token(request)

    @contextlib.contextmanager
    def _snapshot(**kwargs):
        if "token" not in kwargs:
            kwargs["token"] = token
        with _snapshot_context(**kwargs) as snapshot:
            yield snapshot

    return _snapshot


# DEV: The dump_code_to_file function is adapted from the compile function in
# the py_compile module of the Python standard library. It generates .pyc files
# with the right format.
if PY2:
    import marshal
    from py_compile import MAGIC
    from py_compile import wr_long

    def dump_code_to_file(code, file):
        file.write(MAGIC)
        wr_long(file, long(time.time()))  # noqa
        marshal.dump(code, file)
        file.flush()


else:
    import importlib

    code_to_pyc = getattr(
        importlib._bootstrap_external, "_code_to_bytecode" if sys.version_info < (3, 7) else "_code_to_timestamp_pyc"
    )

    def dump_code_to_file(code, file):
        file.write(code_to_pyc(code, time.time(), len(code.co_code)))
        file.flush()


def unwind_params(params):
    if params is None:
        yield None
        return

    for _ in product(*(((k, v) for v in vs) for k, vs in params.items())):
        yield dict(_)


class FunctionDefFinder(ast.NodeVisitor):
    def __init__(self, func_name):
        super(FunctionDefFinder, self).__init__()
        self.func_name = func_name
        self._body = None

    def generic_visit(self, node):
        return self._body or super(FunctionDefFinder, self).generic_visit(node)

    def visit_FunctionDef(self, node):
        if node.name == self.func_name:
            self._body = node.body

    def find(self, file):
        with open(file) as f:
            t = ast.parse(f.read())
            self.visit(t)
            t.body = self._body
            return t


def run_function_from_file(item, params=None):
    file, _, func = item.location
    marker = item.get_closest_marker("subprocess")
    run_module = marker.kwargs.get("run_module", False)

    args = [sys.executable]

    # Add ddtrace-run prefix in ddtrace-run mode
    if marker.kwargs.get("ddtrace_run", False):
        args.insert(0, "ddtrace-run")

    # Add -m if running script as a module
    if run_module:
        args.append("-m")

    # Override environment variables for the subprocess
    env = os.environ.copy()
    env.update(marker.kwargs.get("env", {}))
    if params is not None:
        env.update(params)

    expected_status = marker.kwargs.get("status", 0)

    expected_out = marker.kwargs.get("out", "")
    if expected_out is not None:
        expected_out = expected_out.encode("utf-8")

    expected_err = marker.kwargs.get("err", "")
    if expected_err is not None:
        expected_err = expected_err.encode("utf-8")

    with NamedTemporaryFile(mode="wb", suffix=".pyc") as fp:
        dump_code_to_file(compile(FunctionDefFinder(func).find(file), file, "exec"), fp.file)

        # If running a module with -m, we change directory to the module's
        # folder and run the module directly.
        if run_module:
            cwd, module = split(splitext(fp.name)[0])
            args.append(module)
        else:
            cwd = None
            args.append(fp.name)

        # Add any extra requested args
        args.extend(marker.kwargs.get("args", []))

        def _subprocess_wrapper():
            out, err, status, _ = call_program(*args, env=env, cwd=cwd)

            if status != expected_status:
                raise AssertionError(
                    "Expected status %s, got %s."
                    "\n=== Captured STDOUT ===\n%s=== End of captured STDOUT ==="
                    "\n=== Captured STDERR ===\n%s=== End of captured STDERR ==="
                    % (expected_status, status, out.decode("utf-8"), err.decode("utf-8"))
                )
            elif expected_out is not None and out != expected_out:
                raise AssertionError("STDOUT: Expected [%s] got [%s]" % (expected_out, out))
            elif expected_err is not None and err != expected_err:
                raise AssertionError("STDERR: Expected [%s] got [%s]" % (expected_err, err))

        return TestReport.from_item_and_call(item, CallInfo.from_call(_subprocess_wrapper, "call"))


@pytest.hookimpl(tryfirst=True)
def pytest_runtest_protocol(item):
    marker = item.get_closest_marker("subprocess")
    if marker:
        params = marker.kwargs.get("parametrize", None)
        ihook = item.ihook
        base_name = item.nodeid

        for ps in unwind_params(params):
            nodeid = (base_name + str(ps)) if ps is not None else base_name

            # Start
            ihook.pytest_runtest_logstart(nodeid=nodeid, location=item.location)

            # Setup
            report = call_and_report(item, "setup", log=False)
            report.nodeid = nodeid
            ihook.pytest_runtest_logreport(report=report)

            # Call
            report = run_function_from_file(item, ps)
            report.nodeid = nodeid
            ihook.pytest_runtest_logreport(report=report)

            # Teardown
            report = call_and_report(item, "teardown", log=False, nextitem=None)
            report.nodeid = nodeid
            ihook.pytest_runtest_logreport(report=report)

            # Finish
            ihook.pytest_runtest_logfinish(nodeid=nodeid, location=item.location)

        return True



import base64
import json
import os
from typing import Any
from typing import Generator
from typing import Tuple

import attr
import pytest

from ddtrace.internal.compat import httplib
from ddtrace.internal.compat import parse
from ddtrace.internal.telemetry.writer import TelemetryWriter
from ddtrace.internal.utils.formats import parse_tags_str
from tests.utils import request_token


@pytest.fixture
def telemetry_writer():
    telemetry_writer = TelemetryWriter()
    # Enable the TelemetryWriter without queuing an app-started event
    # and setting up exit hooks
    telemetry_writer._enabled = True
    yield telemetry_writer


@attr.s
class TelemetryTestSession(object):
    token = attr.ib(type=str)
    telemetry_writer = attr.ib(type=TelemetryWriter)

    def create_connection(self):
        parsed = parse.urlparse(self.telemetry_writer._agent_url)
        return httplib.HTTPConnection(parsed.hostname, parsed.port)

    def _request(self, method, url):
        # type: (str, str) -> Tuple[int, bytes]
        conn = self.create_connection()
        try:
            conn.request(method, url)
            r = conn.getresponse()
            return r.status, r.read()
        finally:
            conn.close()

    def clear(self):
        status, _ = self._request("GET", "/test/session/clear?test_session_token=%s" % self.token)
        if status != 200:
            pytest.fail("Failed to clear session: %s" % self.token)
        return True

    def get_requests(self):
        """Get a list of the requests sent to the test agent

        Results are in reverse order by ``seq_id``
        """
        status, body = self._request("GET", "/test/session/requests?test_session_token=%s" % self.token)
        if status != 200:
            pytest.fail("Failed to fetch session requests: %s" % self.token)
        requests = json.loads(body.decode("utf-8"))
        for req in requests:
            body_str = base64.b64decode(req["body"]).decode("utf-8")
            req["body"] = json.loads(body_str)

        return sorted(requests, key=lambda r: r["body"]["seq_id"], reverse=True)

    def get_events(self):
        """Get a list of the event payloads sent to the test agent

        Results are in reverse order by ``seq_id``
        """
        status, body = self._request("GET", "/test/session/apmtelemetry?test_session_token=%s" % self.token)
        if status != 200:
            pytest.fail("Failed to fetch session events: %s" % self.token)
        return sorted(json.loads(body.decode("utf-8")), key=lambda e: e["seq_id"], reverse=True)


@pytest.fixture
def test_agent_session(telemetry_writer, request):
    # type: (TelemetryWriter, Any) -> Generator[TelemetryTestSession, None, None]
    token = request_token(request)

    telemetry_writer._headers["X-Datadog-Test-Session-Token"] = token

    # Also add a header to the environment for subprocesses test cases that might use snapshotting.
    existing_headers = parse_tags_str(os.environ.get("_DD_TELEMETRY_WRITER_ADDITIONAL_HEADERS", ""))
    existing_headers.update({"X-Datadog-Test-Session-Token": token})
    os.environ["_DD_TELEMETRY_WRITER_ADDITIONAL_HEADERS"] = ",".join(
        ["%s:%s" % (k, v) for k, v in existing_headers.items()]
    )

    requests = TelemetryTestSession(token=token, telemetry_writer=telemetry_writer)

    conn = requests.create_connection()
    try:
        conn.request("GET", "/test/session/start?test_session_token=%s" % token)
        r = conn.getresponse()
        print(r.read())
    finally:
        conn.close()

    try:
        yield requests
    finally:
        # Force a flush
        telemetry_writer.periodic()
        del telemetry_writer._headers["X-Datadog-Test-Session-Token"]
        del os.environ["_DD_TELEMETRY_WRITER_ADDITIONAL_HEADERS"]



from collections import deque
import time
from typing import Tuple

import six


if six.PY3:
    from typing import Mapping
    from typing import Sequence
else:
    from collections import Mapping, Sequence

from _libddwaf cimport DDWAF_OBJ_TYPE
from _libddwaf cimport ddwaf_config
from _libddwaf cimport ddwaf_context
from _libddwaf cimport ddwaf_context_destroy
from _libddwaf cimport ddwaf_context_init
from _libddwaf cimport ddwaf_destroy
from _libddwaf cimport ddwaf_get_version
from _libddwaf cimport ddwaf_handle
from _libddwaf cimport ddwaf_init
from _libddwaf cimport ddwaf_object
from _libddwaf cimport ddwaf_required_addresses
from _libddwaf cimport ddwaf_result
from _libddwaf cimport ddwaf_result_free
from _libddwaf cimport ddwaf_ruleset_info
from _libddwaf cimport ddwaf_run
from _libddwaf cimport ddwaf_version
from cpython.bytes cimport PyBytes_AsString
from cpython.bytes cimport PyBytes_Size
from cpython.exc cimport PyErr_Clear
from cpython.exc cimport PyErr_Occurred
from cpython.mem cimport PyMem_Free
from cpython.mem cimport PyMem_Realloc
from cpython.unicode cimport PyUnicode_AsEncodedString
from libc.stdint cimport uint32_t
from libc.stdint cimport uint64_t
from libc.string cimport memset


DEFAULT_DDWAF_TIMEOUT_MS=20


cdef extern from "Python.h":
    const char* PyUnicode_AsUTF8AndSize(object o, Py_ssize_t *size)


def version():
    # type: () -> Tuple[int, int, int]
    """Get the version of libddwaf."""
    cdef ddwaf_version version
    ddwaf_get_version(&version)
    return (version.major, version.minor, version.patch)


cdef inline object _string_to_bytes(object string, const char **ptr, ssize_t *length):
    ptr[0] = NULL
    if isinstance(string, six.binary_type):
        ptr[0] = PyBytes_AsString(string)
        length[0] = PyBytes_Size(string)
        return string
    elif isinstance(string, six.text_type):
        IF PY_MAJOR_VERSION >= 3:
            ptr[0] = PyUnicode_AsUTF8AndSize(string, length)
            if ptr[0] == NULL and PyErr_Occurred():
                # ignore exception from this function as we fallback to
                # PyUnicode_AsEncodedString
                PyErr_Clear()
        if ptr[0] == NULL:
            string = PyUnicode_AsEncodedString(string, "utf-8", "surrogatepass")
            ptr[0] = PyBytes_AsString(string)
            length[0] = PyBytes_Size(string)
        return string
    raise RuntimeError


cdef class _Wrapper(object):
    """
    Wrapper to convert Python objects to ddwaf objects.

    libddwaf represents scalar and composite values using ddwaf objects. This
    wrapper converts Python objects to ddwaf objects by traversing all values
    and their children. By default, the number of objects is limited to avoid
    infinite loops. This limitation can be lifted on trusted data by setting
    `max_objects` to `None`.

    Under the hood, the wrapper uses an array of `ddwaf_object` allocated as a
    single buffer. Objects such as maps or arrays refer to other objects that
    are only part of this buffer. Strings are not copied, they live in the
    Python heap and are referenced by the wrapper to avoid garbage collection.
    """

    cdef ddwaf_object *_ptr
    cdef readonly object _string_refs
    cdef readonly ssize_t _size
    cdef readonly ssize_t _next_idx

    def __init__(self, value, max_objects=5000):
        self._string_refs = []
        self._convert(value, max_objects)

    cdef ssize_t _reserve_obj(self, ssize_t n=1) except -1:
        """
        Exponentially grows the size of the memory space used for objects.
        Will stop if too much memory is allocated.
        """
        cdef ssize_t idx, i
        cdef ddwaf_object *ptr
        cdef ddwaf_object *obj

        idx = self._next_idx
        if idx + n > self._size:
            while idx + n > self._size:
                # grow 1.5 the previous size + an initial fixed size until
                #  it can accommodate at least n new objects
                self._size += (self._size >> 1) + 128
            ptr = <ddwaf_object *> PyMem_Realloc(self._ptr, self._size * sizeof(ddwaf_object))
            if ptr == NULL:
                raise MemoryError
            memset(ptr + idx, 0, (self._size - idx) * sizeof(ddwaf_object))
            if self._ptr != NULL and ptr != self._ptr:
                # we need to patch all array objects because they use pointers to other objects
                for i in range(idx):
                    obj = ptr + i
                    if (obj.type == DDWAF_OBJ_TYPE.DDWAF_OBJ_MAP or obj.type == DDWAF_OBJ_TYPE.DDWAF_OBJ_ARRAY) and obj.array != NULL:
                        obj.array = obj.array - self._ptr + ptr
            self._ptr = ptr
        self._next_idx += n
        return idx

    cdef int _make_string(self, ssize_t idx, object string) except -1:
        cdef const char * ptr
        cdef ssize_t length
        cdef ddwaf_object *obj

        self._string_refs.append(_string_to_bytes(string, &ptr, &length))

        obj = self._ptr + idx
        obj.type = DDWAF_OBJ_TYPE.DDWAF_OBJ_STRING
        obj.stringValue = ptr
        obj.nbEntries = length

    cdef void _make_array(self, ssize_t idx, ssize_t array_idx, ssize_t nb_entries):
        cdef ddwaf_object *obj
        obj = self._ptr + idx
        obj.type = DDWAF_OBJ_TYPE.DDWAF_OBJ_ARRAY
        obj.array = self._ptr + array_idx
        obj.nbEntries = nb_entries

    cdef void _make_map(self, ssize_t idx, ssize_t array_idx, ssize_t nb_entries):
        cdef ddwaf_object *obj
        obj = self._ptr + idx
        obj.type = DDWAF_OBJ_TYPE.DDWAF_OBJ_MAP
        obj.array = self._ptr + array_idx
        obj.nbEntries = nb_entries

    cdef int _set_parameter(self, ssize_t idx, object string) except -1:
        cdef const char * ptr
        cdef ssize_t length
        cdef ddwaf_object *obj

        self._string_refs.append(_string_to_bytes(string, &ptr, &length))

        obj = self._ptr + idx
        obj.parameterName = ptr
        obj.parameterNameLength = length

    cdef void _convert(self, value, max_objects) except *:
        cdef object stack
        cdef ssize_t i, j, n, idx, items_idx

        i = 0
        stack = deque([(self._reserve_obj(), value)], maxlen=max_objects)
        while len(stack) and (max_objects is None or i < <ssize_t?> max_objects):
            idx, val = stack.popleft()

            if isinstance(val, (int, float)):
                val = str(val)

            if isinstance(val, (six.binary_type, six.text_type)):
                self._make_string(idx, val)

            elif isinstance(val, Mapping):
                n = len(val)
                items_idx = self._reserve_obj(n)
                self._make_map(idx, items_idx, n)
                # size of val must not change!! should not happen
                # while holding the GIL?
                for j, (k, v) in enumerate(six.iteritems(val)):
                    if not isinstance(k, (six.binary_type, six.text_type)):
                        if isinstance(k, (int, float)):
                            k = str(k)
                        else:
                            continue
                    self._set_parameter(items_idx + j, k)
                    stack.append((items_idx + j, v))

            elif isinstance(val, Sequence):
                n = len(val)
                items_idx = self._reserve_obj(n)
                self._make_array(idx, items_idx, n)
                stack.extend([(items_idx + j, val[j]) for j in range(n)])

            i += 1

    def __repr__(self):
        return "<{0.__class__.__name__} for {0._next_idx} elements>".format(self)

    def __sizeof__(self):
        return super(_Wrapper, self).__sizeof__() + self._size * sizeof(ddwaf_object)

    def __dealloc__(self):
        PyMem_Free(self._ptr)

cdef str _char_to_str(const char* char_value):
    value = ""
    IF PY_MAJOR_VERSION >= 3:
        value = str(char_value, encoding="utf-8")
    ELSE:
        value = str(char_value)
    return value

cdef class DDWafObject(object):
    cdef ddwaf_object _obj

    cdef init(self, ddwaf_object obj):
        self._obj = obj

    cdef uint64_t length(self):
        return self._obj.nbEntries

    cdef ddwaf_object index_OLD(self, uint64_t i):
        if i == 0:
            return self._obj
        cdef ddwaf_object* base = self._obj.array
        return (<ddwaf_object*>(base + ((i - 1) * sizeof(ddwaf_object))))[0]



    cdef get_value(self, ddwaf_object dd_obj):
        value = ""
        cdef DDWAF_OBJ_TYPE type_ob = dd_obj.type

        if type_ob == DDWAF_OBJ_TYPE.DDWAF_OBJ_STRING:
            value = _char_to_str(dd_obj.stringValue)
        elif type_ob == DDWAF_OBJ_TYPE.DDWAF_OBJ_SIGNED:
            value = int(dd_obj.intValue)
        elif type_ob == DDWAF_OBJ_TYPE.DDWAF_OBJ_UNSIGNED:
            value = int(dd_obj.uintValue)
        elif type_ob == DDWAF_OBJ_TYPE.DDWAF_OBJ_ARRAY:
            value = []
            for i in range(dd_obj.nbEntries):
                result = self._index(dd_obj, i)
                value.append(result)
        elif type_ob == DDWAF_OBJ_TYPE.DDWAF_OBJ_MAP:
            value = {}
            for i in range(dd_obj.nbEntries):
                result = self._index(dd_obj, i)
                value.update(result)
        return value

    cdef _index(self, ddwaf_object obj, uint64_t i):
        cdef ddwaf_object base = obj.array[i]
        value = self.get_value(base)

        if base.parameterName != NULL:
            result = {}
            result[_char_to_str(base.parameterName)] = value
        else:
            result = value
        return result

    cdef index(self, uint64_t i):
        return self._index(self._obj, i)

cdef class DDWaf(object):
    """
    A DDWaf instance performs a matching operation on provided data according
    to some rules.
    """

    cdef ddwaf_handle _handle
    cdef ddwaf_ruleset_info _info
    cdef object _rules

    def __init__(self, rules, obfuscation_parameter_key_regexp, obfuscation_parameter_value_regexp):
        cdef ddwaf_object* rule_objects
        cdef ddwaf_config config

        config = {
            # Default limits
            'limits': {
                "max_container_size": 0,
                "max_container_depth": 0,
                "max_string_length": 0
            },
            'obfuscator': {
                "key_regex": obfuscation_parameter_key_regexp,
                "value_regex": obfuscation_parameter_value_regexp
            },
        }

        self._rules = _Wrapper(rules, max_objects=None)
        rule_objects = (<_Wrapper?>self._rules)._ptr;
        self._handle = ddwaf_init(rule_objects, &config, &self._info)
        if <void *> self._handle == NULL:
            raise ValueError("invalid rules")

    @property
    def required_data(self):
        cdef uint32_t size
        cdef const char* const* ptr

        addresses = []
        ptr = ddwaf_required_addresses(self._handle, &size)
        for i in range(size):
            addresses.append((<bytes> ptr[i]).decode("utf-8"))
        return addresses

    @property
    def info(self):
        cdef dict result
        cdef dict errors_result = {}
        cdef DDWafObject errors_ob = DDWafObject()

        errors_ob.init(self._info.errors)

        if self._info.loaded > 0:
            if self._info.failed > 0:
                for i in range(errors_ob.length()):
                    result = errors_ob.index(<uint64_t?>i)
                    errors_result.update(result)

            version = ""
            if self._info.version != NULL:
                version = _char_to_str(self._info.version)
            return {
                "loaded": self._info.loaded,
                "failed": self._info.failed,
                "errors": errors_result,
                "version": version,
            }
        return {
                "loaded": 0,
                "failed": 0,
                "errors": errors_result,
                "version": "",
            }

    def run(self, data, timeout_ms=DEFAULT_DDWAF_TIMEOUT_MS):
        start = time.time()
        cdef ddwaf_context ctx
        cdef ddwaf_result result

        ctx = ddwaf_context_init(self._handle, NULL)
        if <void *> ctx == NULL:
            raise RuntimeError
        try:
            wrapper = _Wrapper(data)
            ddwaf_run(ctx, (<_Wrapper?>wrapper)._ptr, &result, <uint64_t?> timeout_ms * 1000)
            if result.data != NULL:
                return (<bytes> result.data).decode("utf-8"), result.total_runtime / 1e3, (time.time() - start) * 1e6
            return None, result.total_runtime / 1e3, (time.time() - start) * 1e6
        finally:
            ddwaf_result_free(&result)
            ddwaf_context_destroy(ctx)

    def __dealloc__(self):
        ddwaf_destroy(self._handle)



import logging

import pytest
import six

import ddtrace
from ddtrace.constants import ENV_KEY
from ddtrace.constants import VERSION_KEY
from ddtrace.contrib.logging import patch
from ddtrace.contrib.logging import unpatch
from ddtrace.contrib.logging.patch import RECORD_ATTR_SPAN_ID
from ddtrace.contrib.logging.patch import RECORD_ATTR_TRACE_ID
from ddtrace.internal.compat import StringIO
from ddtrace.vendor import wrapt
from tests.utils import TracerTestCase


logger = logging.getLogger()
logger.level = logging.INFO

DEFAULT_FORMAT = (
    "%(message)s - dd.service=%(dd.service)s dd.version=%(dd.version)s dd.env=%(dd.env)s"
    " dd.trace_id=%(dd.trace_id)s dd.span_id=%(dd.span_id)s"
)


def current_span(tracer=None):
    if not tracer:
        tracer = ddtrace.tracer
    return tracer.current_span()


class AssertFilter(logging.Filter):
    def filter(self, record):
        trace_id = getattr(record, RECORD_ATTR_TRACE_ID)
        assert isinstance(trace_id, six.string_types)

        span_id = getattr(record, RECORD_ATTR_SPAN_ID)
        assert isinstance(span_id, six.string_types)

        return True


def capture_function_log(func, fmt=DEFAULT_FORMAT, logger_override=None, fmt_style=None):
    if logger_override is not None:
        logger_to_capture = logger_override
    else:
        logger_to_capture = logger

    # add stream handler to capture output
    out = StringIO()
    sh = logging.StreamHandler(out)

    try:
        if fmt_style:
            formatter = logging.Formatter(fmt, style=fmt_style)
        else:
            formatter = logging.Formatter(fmt)
        sh.setFormatter(formatter)
        logger_to_capture.addHandler(sh)
        assert_filter = AssertFilter()
        logger_to_capture.addFilter(assert_filter)
        result = func()
    finally:
        logger_to_capture.removeHandler(sh)
        logger_to_capture.removeFilter(assert_filter)

    return out.getvalue().strip(), result


class LoggingTestCase(TracerTestCase):
    def setUp(self):
        patch()
        super(LoggingTestCase, self).setUp()

    def tearDown(self):
        unpatch()
        super(LoggingTestCase, self).tearDown()

    def test_patch(self):
        """
        Confirm patching was successful
        """
        log = logging.getLogger()
        self.assertTrue(isinstance(log.makeRecord, wrapt.BoundFunctionWrapper))
        # For Python 3
        if hasattr(logging, "StrFormatStyle"):
            if hasattr(logging.StrFormatStyle, "_format"):
                assert isinstance(logging.StrFormatStyle._format, wrapt.BoundFunctionWrapper)
            else:
                assert isinstance(logging.StrFormatStyle.format, wrapt.BoundFunctionWrapper)

        unpatch()
        log = logging.getLogger()
        self.assertFalse(isinstance(log.makeRecord, wrapt.BoundFunctionWrapper))
        # For Python 3
        if hasattr(logging, "StrFormatStyle"):
            if hasattr(logging.StrFormatStyle, "_format"):
                assert not isinstance(logging.StrFormatStyle._format, wrapt.BoundFunctionWrapper)
            else:
                assert not isinstance(logging.StrFormatStyle.format, wrapt.BoundFunctionWrapper)

    def _test_logging(self, create_span, service="", version="", env=""):
        def func():
            span = create_span()
            logger.info("Hello!")
            if span:
                span.finish()
            return span

        with self.override_config("logging", dict(tracer=self.tracer)):
            # with format string for trace info
            output, span = capture_function_log(func)
            trace_id = 0
            span_id = 0
            if span:
                trace_id = span.trace_id
                span_id = span.span_id

            assert output.startswith(
                "Hello! - dd.service={} dd.version={} dd.env={} dd.trace_id={} dd.span_id={}".format(
                    service, version, env, trace_id, span_id
                )
            )

            # without format string
            output, _ = capture_function_log(func, fmt="%(message)s")
            assert output == "Hello!"

    def test_log_trace(self):
        """
        Check logging patched and formatter including trace info
        """

        def create_span():
            return self.tracer.trace("test.logging")

        self._test_logging(create_span=create_span)

        with self.override_global_config(dict(version="global.version", env="global.env")):
            self._test_logging(create_span=create_span, version="global.version", env="global.env")

    def test_log_trace_service(self):
        def create_span():
            return self.tracer.trace("test.logging", service="logging")

        self._test_logging(create_span=create_span)

        with self.override_global_config(dict(version="global.version", env="global.env")):
            self._test_logging(create_span=create_span, version="global.version", env="global.env")

    @TracerTestCase.run_in_subprocess(env_overrides=dict(DD_TAGS="service:ddtagservice,env:ddenv,version:ddversion"))
    def test_log_DD_TAGS(self):
        def create_span():
            return self.tracer.trace("test.logging")

        self._test_logging(create_span=create_span, service="ddtagservice", version="ddversion", env="ddenv")

    def test_log_trace_version(self):
        def create_span():
            span = self.tracer.trace("test.logging")
            span.set_tag(VERSION_KEY, "manual.version")
            return span

        self._test_logging(create_span=create_span, version="")

        # Setting global config version and overriding with span specific value
        # We always want the globals in the logs
        with self.override_global_config(dict(version="global.version", env="global.env")):
            self._test_logging(create_span=create_span, version="global.version", env="global.env")

    def test_log_trace_env(self):
        """
        Check logging patched and formatter including trace info
        """

        def create_span():
            span = self.tracer.trace("test.logging")
            span.set_tag(ENV_KEY, "manual.env")
            return span

        self._test_logging(create_span=create_span, env="")

        # Setting global config env and overriding with span specific value
        # We always want the globals in the logs
        with self.override_global_config(dict(version="global.version", env="global.env")):
            self._test_logging(create_span=create_span, version="global.version", env="global.env")

    def test_log_no_trace(self):
        """
        Check traced funclogging patched and formatter not including trace info
        """

        def create_span():
            return None

        self._test_logging(create_span=create_span)

        with self.override_global_config(dict(version="global.version", env="global.env")):
            self._test_logging(create_span=create_span, version="global.version", env="global.env")

    @pytest.mark.skipif(six.PY2, reason="logging.StrFormatStyle does not exist on Python 2.7")
    def test_log_strformat_style(self):
        def func():
            with self.tracer.trace("test.logging") as span:
                logger.info("Hello!")
                return span

        fmt = (
            "{message} [dd.service={dd.service} dd.env={dd.env} "
            "dd.version={dd.version} dd.trace_id={dd.trace_id} dd.span_id={dd.span_id}]"
        )

        with self.override_config("logging", dict(tracer=self.tracer)):
            output, span = capture_function_log(func, fmt=fmt, fmt_style="{")

            lines = output.splitlines()
            assert (
                "Hello! [dd.service= dd.env= dd.version= dd.trace_id={} dd.span_id={}]".format(
                    span.trace_id, span.span_id
                )
                == lines[0]
            )

            with self.override_global_config(dict(service="my.service", version="my.version", env="my.env")):
                output, span = capture_function_log(func, fmt=fmt, fmt_style="{")

                lines = output.splitlines()
                expected = (
                    "Hello! [dd.service=my.service dd.env=my.env dd.version=my.version dd.trace_id={} dd.span_id={}]"
                ).format(span.trace_id, span.span_id)
                assert expected == lines[0]

    @pytest.mark.skipif(six.PY2, reason="logging.StrFormatStyle does not exist on Python 2.7")
    def test_log_strformat_style_format(self):
        # DEV: We have to use `{msg}` instead of `{message}` because we are manually creating
        # records which does not properly configure `record.message` attribute
        fmt = (
            "{msg} [dd.service={dd.service} dd.env={dd.env} "
            "dd.version={dd.version} dd.trace_id={dd.trace_id} dd.span_id={dd.span_id}]"
        )
        formatter = logging.StrFormatStyle(fmt)

        with self.override_config("logging", dict(tracer=self.tracer)):
            with self.tracer.trace("test.logging") as span:
                record = logger.makeRecord("name", "INFO", "func", 534, "Manual log record", (), None)
                log = formatter.format(record)
                expected = "Manual log record [dd.service= dd.env= dd.version= dd.trace_id={} dd.span_id={}]".format(
                    span.trace_id, span.span_id
                )
                assert log == expected

                assert not hasattr(record, "dd")
                assert getattr(record, RECORD_ATTR_TRACE_ID) == str(span.trace_id)
                assert getattr(record, RECORD_ATTR_SPAN_ID) == str(span.span_id)



import os
import typing as t

from debugger import COLS
from debugger import ExplorationDebugger
from debugger import ModuleCollector
from debugger import status

from ddtrace.debugging._function.discovery import FunctionDiscovery
from ddtrace.debugging._probe.model import FunctionProbe
from ddtrace.internal.utils.formats import asbool


# Track all instrumented functions and their call count.
_tracked_funcs = {}  # type: t.Dict[str, int]


ENABLED = asbool(os.getenv("DD_DEBUGGER_EXPL_PROFILER_ENABLED", True))


class FunctionCollector(ModuleCollector):
    def on_collect(self, discovery):
        # type: (FunctionDiscovery) -> None
        module = discovery._module
        status("[profiler] Collecting functions from %s" % module.__name__)
        for fname, f in discovery._fullname_index.items():
            _tracked_funcs[fname] = 0
            DeterministicProfiler.add_probe(
                FunctionProbe(
                    probe_id=str(hash(f)),
                    module=module.__name__,
                    func_qname=fname.replace(module.__name__, "").lstrip("."),
                    rate=float("inf"),
                )
            )


class DeterministicProfiler(ExplorationDebugger):
    __watchdog__ = FunctionCollector

    @classmethod
    def report_func_calls(cls):
        # type: () -> None
        for probe in (_ for _ in cls.get_triggered_probes() if isinstance(_, FunctionProbe)):
            _tracked_funcs[".".join([probe.module, probe.func_qname])] += 1
        print(("{:=^%ds}" % COLS).format(" Function coverage "))
        print("")
        calls = sorted([(v, k) for k, v in _tracked_funcs.items()], reverse=True)
        if not calls:
            print("No functions called")
            return
        w = max(len(f) for _, f in calls)
        called = sum(v > 0 for v in _tracked_funcs.values())
        print("Functions called: %d/%d" % (called, len(_tracked_funcs)))
        print("")
        print(("{:<%d} {:>5}" % w).format("Function", "Calls"))
        print("=" * (w + 6))
        for calls, func in calls:
            print(("{:<%d} {:>5}" % w).format(func, calls))
        print("")

    @classmethod
    def on_disable(cls):
        # type: () -> None
        cls.report_func_calls()


if ENABLED:
    DeterministicProfiler.enable()



from ddtrace import config

from .. import trace_utils
from ...constants import ANALYTICS_SAMPLE_RATE_KEY
from ...constants import SPAN_MEASURED_KEY
from ...ext import SpanTypes
from ...ext import http
from ...internal.compat import stringify
from ..asyncio import context_provider


CONFIG_KEY = "datadog_trace"
REQUEST_CONTEXT_KEY = "datadog_context"
REQUEST_CONFIG_KEY = "__datadog_trace_config"
REQUEST_SPAN_KEY = "__datadog_request_span"


async def trace_middleware(app, handler):
    """
    ``aiohttp`` middleware that traces the handler execution.
    Because handlers are run in different tasks for each request, we attach the Context
    instance both to the Task and to the Request objects. In this way:

    * the Task is used by the internal automatic instrumentation
    * the ``Context`` attached to the request can be freely used in the application code
    """

    async def attach_context(request):
        # application configs
        tracer = app[CONFIG_KEY]["tracer"]
        service = app[CONFIG_KEY]["service"]
        distributed_tracing = app[CONFIG_KEY]["distributed_tracing_enabled"]
        # Create a new context based on the propagated information.
        trace_utils.activate_distributed_headers(
            tracer,
            int_config=config.aiohttp,
            request_headers=request.headers,
            override=distributed_tracing,
        )

        # trace the handler
        request_span = tracer.trace(
            "aiohttp.request",
            service=service,
            span_type=SpanTypes.WEB,
        )
        request_span.set_tag(SPAN_MEASURED_KEY)

        # Configure trace search sample rate
        # DEV: aiohttp is special case maintains separate configuration from config api
        analytics_enabled = app[CONFIG_KEY]["analytics_enabled"]
        if (config.analytics_enabled and analytics_enabled is not False) or analytics_enabled is True:
            request_span.set_tag(ANALYTICS_SAMPLE_RATE_KEY, app[CONFIG_KEY].get("analytics_sample_rate", True))

        # attach the context and the root span to the request; the Context
        # may be freely used by the application code
        request[REQUEST_CONTEXT_KEY] = request_span.context
        request[REQUEST_SPAN_KEY] = request_span
        request[REQUEST_CONFIG_KEY] = app[CONFIG_KEY]
        try:
            response = await handler(request)
            return response
        except Exception:
            request_span.set_traceback()
            raise

    return attach_context


async def on_prepare(request, response):
    """
    The on_prepare signal is used to close the request span that is created during
    the trace middleware execution.
    """
    # safe-guard: discard if we don't have a request span
    request_span = request.get(REQUEST_SPAN_KEY, None)
    if not request_span:
        return

    # default resource name
    resource = stringify(response.status)

    if request.match_info.route.resource:
        # collect the resource name based on http resource type
        res_info = request.match_info.route.resource.get_info()

        if res_info.get("path"):
            resource = res_info.get("path")
        elif res_info.get("formatter"):
            resource = res_info.get("formatter")
        elif res_info.get("prefix"):
            resource = res_info.get("prefix")

        # prefix the resource name by the http method
        resource = "{} {}".format(request.method, resource)

    request_span.resource = resource

    # DEV: aiohttp is special case maintains separate configuration from config api
    trace_query_string = request[REQUEST_CONFIG_KEY].get("trace_query_string")
    if trace_query_string is None:
        trace_query_string = config.http.trace_query_string
    if trace_query_string:
        request_span.set_tag(http.QUERY_STRING, request.query_string)

    trace_utils.set_http_meta(
        request_span,
        config.aiohttp,
        method=request.method,
        url=str(request.url),  # DEV: request.url is a yarl's URL object
        status_code=response.status,
        request_headers=request.headers,
        response_headers=response.headers,
    )

    request_span.finish()


def trace_app(app, tracer, service="aiohttp-web"):
    """
    Tracing function that patches the ``aiohttp`` application so that it will be
    traced using the given ``tracer``.

    :param app: aiohttp application to trace
    :param tracer: tracer instance to use
    :param service: service name of tracer
    """

    # safe-guard: don't trace an application twice
    if getattr(app, "__datadog_trace", False):
        return
    setattr(app, "__datadog_trace", True)

    # configure datadog settings
    app[CONFIG_KEY] = {
        "tracer": tracer,
        "service": config._get_service(default=service),
        "distributed_tracing_enabled": None,
        "analytics_enabled": None,
        "analytics_sample_rate": 1.0,
    }

    # the tracer must work with asynchronous Context propagation
    tracer.configure(context_provider=context_provider)

    # add the async tracer middleware as a first middleware
    # and be sure that the on_prepare signal is the last one
    app.middlewares.insert(0, trace_middleware)
    app.on_response_prepare.append(on_prepare)



import os
import signal
import subprocess
import time
from typing import Dict
from typing import Generator
from typing import List

import pytest
import tenacity

from ddtrace.contrib.flask.patch import flask_version
from tests.webclient import Client


DEFAULT_HEADERS = {
    "User-Agent": "python-httpx/x.xx.x",
}


@pytest.fixture
def flask_port():
    # type: () -> str
    return "8000"


@pytest.fixture
def flask_wsgi_application():
    # type: () -> str
    return "tests.contrib.flask.app:app"


@pytest.fixture
def flask_command(flask_wsgi_application, flask_port):
    # type: (str, str) -> List[str]
    cmd = "ddtrace-run flask run -h 0.0.0.0 -p %s" % (flask_port,)
    return cmd.split()


@pytest.fixture
def flask_env(flask_wsgi_application):
    # type: (str) -> Dict[str, str]
    env = os.environ.copy()
    env.update(
        {
            # Avoid noisy database spans being output on app startup/teardown.
            "DD_TRACE_SQLITE3_ENABLED": "0",
            "FLASK_APP": flask_wsgi_application,
        }
    )
    return env


@pytest.fixture
def flask_client(flask_command, flask_env, flask_port):
    # type: (str, Dict[str, str], str) -> Generator[Client, None, None]
    # Copy the env to get the correct PYTHONPATH and such
    # from the virtualenv.
    # webservers might exec or fork into another process, so we need to os.setsid() to create a process group
    # (all of which will listen to signals sent to the parent) so that we can kill the whole application.
    proc = subprocess.Popen(
        flask_command,
        stdout=subprocess.PIPE,
        stderr=subprocess.PIPE,
        close_fds=True,
        env=flask_env,
        preexec_fn=os.setsid,
    )
    try:
        client = Client("http://0.0.0.0:%s" % flask_port)
        # Wait for the server to start up
        try:
            client.wait()
        except tenacity.RetryError:
            # if proc.returncode is not None:
            # process failed
            stdout = proc.stdout.read()
            stderr = proc.stderr.read()
            raise TimeoutError(
                "Server failed to start\n======STDOUT=====%s\n\n======STDERR=====%s\n" % (stdout, stderr)
            )
        yield client
        try:
            client.get_ignored("/shutdown")
        except Exception:
            pass
        # At this point the traces have been sent to the test agent
        # but the test agent hasn't necessarily finished processing
        # the traces (race condition) so wait just a bit for that
        # processing to complete.
        time.sleep(0.2)
    finally:
        os.killpg(proc.pid, signal.SIGKILL)
        proc.wait()


@pytest.mark.snapshot(
    ignores=["meta.flask.version"], variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)}
)
def test_flask_200(flask_client):
    # type: (Client) -> None
    assert flask_client.get("/", headers=DEFAULT_HEADERS).status_code == 200


@pytest.mark.snapshot(
    ignores=["meta.flask.version"], variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)}
)
def test_flask_stream(flask_client):
    # type: (Client) -> None
    assert flask_client.get("/stream", headers=DEFAULT_HEADERS, stream=True).status_code == 200


@pytest.mark.snapshot(
    ignores=["meta.flask.version", "meta.http.useragent"],
    variants={"220": flask_version >= (2, 2, 0), "": flask_version < (2, 2, 0)},
)
def test_flask_get_user(flask_client):
    # type: (Client) -> None
    assert flask_client.get("/identify").status_code == 200



import typing
from typing import Any

from ddtrace.profiling import exporter
from ddtrace.profiling import recorder as recorder
from ddtrace.profiling.collector import _lock
from ddtrace.profiling.collector import memalloc
from ddtrace.profiling.collector import stack_event
from ddtrace.profiling.collector import threading as threading

stdlib_path: Any
platstdlib_path: Any
purelib_path: Any
platlib_path: Any
STDLIB: Any

class Package(typing.TypedDict):
    name: str
    version: str
    kind: typing.Literal["library"]
    paths: typing.List[str]

class _Sequence:
    start_at: Any = ...
    next_id: Any = ...
    def __attrs_post_init__(self) -> None: ...
    def generate(self) -> int: ...
    def __init__(self, start_at: Any, next_id: Any) -> None: ...
    def __lt__(self, other: Any) -> Any: ...
    def __le__(self, other: Any) -> Any: ...
    def __gt__(self, other: Any) -> Any: ...
    def __ge__(self, other: Any) -> Any: ...

class _StringTable:
    def to_id(self, string: str) -> int: ...
    def __iter__(self) -> typing.Iterator[str]: ...
    def __len__(self) -> int: ...
    def __init__(self, strings: Any, seq_id: Any) -> None: ...
    def __lt__(self, other: Any) -> Any: ...
    def __le__(self, other: Any) -> Any: ...
    def __gt__(self, other: Any) -> Any: ...
    def __ge__(self, other: Any) -> Any: ...

class pprof_LocationType:
    id: int

class pprof_Mapping:
    filename: int

class pprof_ProfileType:
    id: int
    string_table: typing.Dict[int, str]
    mapping: typing.List[pprof_Mapping]
    def SerializeToString(self) -> bytes: ...

class pprof_FunctionType:
    id: int

HashableStackTraceType: Any

class _PprofConverter:
    def convert_stack_event(
        self,
        thread_id: str,
        thread_native_id: str,
        thread_name: str,
        task_id: str,
        task_name: str,
        local_root_span_id: str,
        span_id: str,
        trace_resource: str,
        trace_type: str,
        frames: HashableStackTraceType,
        nframes: int,
        samples: typing.List[stack_event.StackSampleEvent],
    ) -> None: ...
    def convert_memalloc_event(
        self,
        thread_id: str,
        thread_native_id: str,
        thread_name: str,
        frames: HashableStackTraceType,
        nframes: int,
        events: typing.List[memalloc.MemoryAllocSampleEvent],
    ) -> None: ...
    def convert_memalloc_heap_event(self, event: memalloc.MemoryHeapSampleEvent) -> None: ...
    def convert_lock_acquire_event(
        self,
        lock_name: str,
        thread_id: str,
        thread_name: str,
        task_id: str,
        task_name: str,
        local_root_span_id: str,
        span_id: str,
        trace_resource: str,
        trace_type: str,
        frames: HashableStackTraceType,
        nframes: int,
        events: typing.List[_lock.LockAcquireEvent],
        sampling_ratio: float,
    ) -> None: ...
    def convert_lock_release_event(
        self,
        lock_name: str,
        thread_id: str,
        thread_name: str,
        task_id: str,
        task_name: str,
        local_root_span_id: str,
        span_id: str,
        trace_resource: str,
        trace_type: str,
        frames: HashableStackTraceType,
        nframes: int,
        events: typing.List[_lock.LockReleaseEvent],
        sampling_ratio: float,
    ) -> None: ...
    def convert_stack_exception_event(
        self,
        thread_id: str,
        thread_native_id: str,
        thread_name: str,
        local_root_span_id: str,
        span_id: str,
        trace_resource: str,
        trace_type: str,
        frames: HashableStackTraceType,
        nframes: int,
        exc_type_name: str,
        events: typing.List[stack_event.StackExceptionSampleEvent],
    ) -> None: ...
    def __init__(
        self,
        functions: Any,
        locations: Any,
        string_table: Any,
        last_location_id: Any,
        last_func_id: Any,
        location_values: Any,
    ) -> None: ...
    def __lt__(self, other: Any) -> Any: ...
    def __le__(self, other: Any) -> Any: ...
    def __gt__(self, other: Any) -> Any: ...
    def __ge__(self, other: Any) -> Any: ...

StackEventGroupKey: Any
LockEventGroupKey: Any
StackExceptionEventGroupKey: Any

class PprofExporter(exporter.Exporter):
    def export(
        self, events: recorder.EventsType, start_time_ns: int, end_time_ns: int
    ) -> typing.Tuple[pprof_ProfileType, typing.List[Package]]: ...
    def __init__(self) -> None: ...
    def __lt__(self, other: Any) -> Any: ...
    def __le__(self, other: Any) -> Any: ...
    def __gt__(self, other: Any) -> Any: ...
    def __ge__(self, other: Any) -> Any: ...



import os

from django.http import HttpResponse
from django.urls import path


filepath, extension = os.path.splitext(__file__)
ROOT_URLCONF = os.path.basename(filepath)
DEBUG = False
SECRET_KEY = "fdsfdasfa"
ALLOWED_HOSTS = ["*"]


def index(request):
    return HttpResponse("test")


urlpatterns = [
    path("", index),
]



import pytest

from ddtrace.debugging import DynamicInstrumentation


@pytest.mark.subprocess(ddtrace_run=True, env=dict(DD_DYNAMIC_INSTRUMENTATION_ENABLED="true"), err=None)
def test_debugger_enabled_ddtrace_run():
    from ddtrace.debugging import DynamicInstrumentation

    assert DynamicInstrumentation._instance is not None


@pytest.mark.subprocess(ddtrace_run=True, err=None)
def test_debugger_disabled_ddtrace_run():
    from ddtrace.debugging import DynamicInstrumentation

    assert DynamicInstrumentation._instance is None


def test_debugger_enabled_programmatically():
    DynamicInstrumentation.enable()
    assert DynamicInstrumentation._instance is not None

    DynamicInstrumentation.disable()
    assert DynamicInstrumentation._instance is None


@pytest.mark.subprocess(err=None)
def test_debugger_fork():
    import os
    import sys

    from ddtrace.debugging import DynamicInstrumentation
    from ddtrace.internal.service import ServiceStatus

    DynamicInstrumentation.enable()
    assert DynamicInstrumentation._instance is not None
    assert DynamicInstrumentation._instance.status == ServiceStatus.RUNNING

    parent_instance = DynamicInstrumentation._instance

    child_pid = os.fork()
    if child_pid == 0:
        assert DynamicInstrumentation._instance is not None
        assert DynamicInstrumentation._instance.status == ServiceStatus.RUNNING

        # Proof that the debugger was actually restarted in the child process
        assert DynamicInstrumentation._instance is not parent_instance

        exit(0)

    _, status = os.waitpid(child_pid, 0)
    assert DynamicInstrumentation._instance is not None
    assert DynamicInstrumentation._instance.status == ServiceStatus.RUNNING
    sys.exit(os.WEXITSTATUS(status))



from typing import Any
from typing import Dict
from typing import List
from typing import Optional

import starlette
from starlette.middleware import Middleware
from starlette.routing import Match

from ddtrace import config
from ddtrace.contrib.asgi.middleware import TraceMiddleware
from ddtrace.internal.logger import get_logger
from ddtrace.internal.utils import get_argument_value
from ddtrace.internal.utils.deprecations import DDTraceDeprecationWarning
from ddtrace.internal.utils.wrappers import unwrap as _u
from ddtrace.span import Span
from ddtrace.vendor.debtcollector import deprecate
from ddtrace.vendor.debtcollector import removals
from ddtrace.vendor.wrapt import ObjectProxy
from ddtrace.vendor.wrapt import wrap_function_wrapper as _w


log = get_logger(__name__)

config._add(
    "starlette",
    dict(
        _default_service="starlette",
        request_span_name="starlette.request",
        distributed_tracing=True,
        aggregate_resources=True,
    ),
)


@removals.remove(removal_version="2.0.0", category=DDTraceDeprecationWarning)
def get_resource(scope):
    path = None
    routes = scope["app"].routes
    for route in routes:
        match, _ = route.matches(scope)
        if match == Match.FULL:
            path = route.path
            break
        elif match == Match.PARTIAL and path is None:
            path = route.path
    return path


@removals.remove(removal_version="2.0.0", category=DDTraceDeprecationWarning)
def span_modifier(span, scope):
    resource = get_resource(scope)
    if config.starlette["aggregate_resources"] and resource:
        span.resource = "{} {}".format(scope["method"], resource)


def traced_init(wrapped, instance, args, kwargs):
    mw = kwargs.pop("middleware", [])
    mw.insert(0, Middleware(TraceMiddleware, integration_config=config.starlette))
    kwargs.update({"middleware": mw})

    wrapped(*args, **kwargs)


def patch():
    if getattr(starlette, "_datadog_patch", False):
        return

    setattr(starlette, "_datadog_patch", True)

    _w("starlette.applications", "Starlette.__init__", traced_init)

    # We need to check that Fastapi instrumentation hasn't already patched these
    if not isinstance(starlette.routing.Route.handle, ObjectProxy):
        _w("starlette.routing", "Route.handle", traced_handler)
    if not isinstance(starlette.routing.Mount.handle, ObjectProxy):
        _w("starlette.routing", "Mount.handle", traced_handler)


def unpatch():
    if not getattr(starlette, "_datadog_patch", False):
        return

    setattr(starlette, "_datadog_patch", False)

    _u(starlette.applications.Starlette, "__init__")

    # We need to check that Fastapi instrumentation hasn't already unpatched these
    if isinstance(starlette.routing.Route.handle, ObjectProxy):
        _u(starlette.routing.Route, "handle")

    if isinstance(starlette.routing.Mount.handle, ObjectProxy):
        _u(starlette.routing.Mount, "handle")


def traced_handler(wrapped, instance, args, kwargs):
    if config.starlette.get("aggregate_resources") is False or config.fastapi.get("aggregate_resources") is False:
        deprecate(
            "ddtrace.contrib.starlette.patch",
            message="`aggregate_resources` is deprecated and will be removed in tracer version 2.0.0",
            category=DDTraceDeprecationWarning,
        )

        return wrapped(*args, **kwargs)

    # Since handle can be called multiple times for one request, we take the path of each instance
    # Then combine them at the end to get the correct resource names
    scope = get_argument_value(args, kwargs, 0, "scope")  # type: Optional[Dict[str, Any]]
    if not scope:
        return wrapped(*args, **kwargs)

    # Our ASGI TraceMiddleware has not been called, skip since
    # we won't have a request span to attach this information onto
    # DEV: This can happen if patching happens after the app has been created
    if "datadog" not in scope:
        log.warning("datadog context not present in ASGI request scope, trace middleware may be missing")
        return wrapped(*args, **kwargs)

    # Add the path to the resource_paths list
    if "resource_paths" not in scope["datadog"]:
        scope["datadog"]["resource_paths"] = [instance.path]
    else:
        scope["datadog"]["resource_paths"].append(instance.path)

    request_spans = scope["datadog"].get("request_spans", [])  # type: List[Span]
    resource_paths = scope["datadog"].get("resource_paths", [])  # type: List[str]

    if len(request_spans) == len(resource_paths):
        # Iterate through the request_spans and assign the correct resource name to each
        for index, span in enumerate(request_spans):
            # We want to set the full resource name on the first request span
            # And one part less of the full resource name for each proceeding request span
            # e.g. full path is /subapp/hello/{name}, first request span gets that as resource name
            # Second request span gets /hello/{name}
            path = "".join(resource_paths[index:])

            if scope.get("method"):
                span.resource = "{} {}".format(scope["method"], path)
            else:
                span.resource = path
    # at least always update the root asgi span resource name request_spans[0].resource = "".join(resource_paths)
    elif request_spans and resource_paths:
        if scope.get("method"):
            request_spans[0].resource = "{} {}".format(scope["method"], "".join(resource_paths))
        else:
            request_spans[0].resource = "".join(resource_paths)

    else:
        log.debug(
            "unable to update the request span resource name, request_spans:%r, resource_paths:%r",
            request_spans,
            resource_paths,
        )

    return wrapped(*args, **kwargs)



import yaaredis

from ddtrace import config
from ddtrace.vendor import wrapt

from ...internal.utils.formats import stringify_cache_args
from ...internal.utils.wrappers import unwrap
from ...pin import Pin
from ..redis.util import _trace_redis_cmd
from ..redis.util import _trace_redis_execute_pipeline


config._add("yaaredis", dict(_default_service="redis"))


def patch():
    """Patch the instrumented methods"""
    if getattr(yaaredis, "_datadog_patch", False):
        return
    setattr(yaaredis, "_datadog_patch", True)

    _w = wrapt.wrap_function_wrapper

    _w("yaaredis.client", "StrictRedis.execute_command", traced_execute_command)
    _w("yaaredis.client", "StrictRedis.pipeline", traced_pipeline)
    _w("yaaredis.pipeline", "StrictPipeline.execute", traced_execute_pipeline)
    _w("yaaredis.pipeline", "StrictPipeline.immediate_execute_command", traced_execute_command)
    Pin().onto(yaaredis.StrictRedis)


def unpatch():
    if getattr(yaaredis, "_datadog_patch", False):
        setattr(yaaredis, "_datadog_patch", False)

        unwrap(yaaredis.client.StrictRedis, "execute_command")
        unwrap(yaaredis.client.StrictRedis, "pipeline")
        unwrap(yaaredis.pipeline.StrictPipeline, "execute")
        unwrap(yaaredis.pipeline.StrictPipeline, "immediate_execute_command")


async def traced_execute_command(func, instance, args, kwargs):
    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return await func(*args, **kwargs)

    with _trace_redis_cmd(pin, config.yaaredis, instance, args):
        return await func(*args, **kwargs)


async def traced_pipeline(func, instance, args, kwargs):
    pipeline = await func(*args, **kwargs)
    pin = Pin.get_from(instance)
    if pin:
        pin.onto(pipeline)
    return pipeline


async def traced_execute_pipeline(func, instance, args, kwargs):
    pin = Pin.get_from(instance)
    if not pin or not pin.enabled():
        return await func(*args, **kwargs)

    cmds = [stringify_cache_args(c) for c, _ in instance.command_stack]
    resource = "\n".join(cmds)
    with _trace_redis_execute_pipeline(pin, config.yaaredis, resource, instance):
        return await func(*args, **kwargs)



from __future__ import absolute_import

import threading
import typing
import weakref

import attr

from ddtrace.internal import nogevent


cpdef get_thread_name(thread_id):
    # This is a special case for gevent:
    # When monkey patching, gevent replaces all active threads by their greenlet equivalent.
    # This means there's no chance to find the MainThread in the list of _active threads.
    # Therefore we special case the MainThread that way.
    # If native threads are started using gevent.threading, they will be inserted in threading._active
    # so we will find them normally.
    if thread_id == nogevent.main_thread_id:
        return "MainThread"

    # We don't want to bother to lock anything here, especially with eventlet involved 😓. We make a best effort to
    # get the thread name; if we fail, it'll just be an anonymous thread because it's either starting or dying.
    try:
        return threading._active[thread_id].name
    except KeyError:
        try:
            return threading._limbo[thread_id].name
        except KeyError:
            return None


cpdef get_thread_native_id(thread_id):
    try:
        thread_obj = threading._active[thread_id]
    except KeyError:
        # This should not happen, unless somebody started a thread without
        # using the `threading` module.
        # In that case, well… just use the thread_id as native_id 🤞
        return thread_id
    else:
        # We prioritize using native ids since we expect them to be surely unique for a program. This is less true
        # for hashes since they are relative to the memory address which can easily be the same across different
        # objects.
        try:
            return thread_obj.native_id
        except AttributeError:
            # Python < 3.8
            return hash(thread_obj)


# cython does not play well with mypy
if typing.TYPE_CHECKING:
    _T = typing.TypeVar("_T")
    _thread_link_base = typing.Generic[_T]
    _weakref_type = weakref.ReferenceType[_T]
else:
    _thread_link_base = object
    _weakref_type = typing.Any


@attr.s(slots=True, eq=False)
class _ThreadLink(_thread_link_base):
    """Link a thread with an object.

    Object is removed when the thread disappears.
    """

    # Key is a thread_id
    # Value is a weakref to an object
    _thread_id_to_object = attr.ib(factory=dict, repr=False, init=False, type=typing.Dict[int, _weakref_type])

    def link_object(
            self,
            obj  # type: _T
    ):
        # type: (...) -> None
        """Link an object to the current running thread."""
        self._thread_id_to_object[nogevent.thread_get_ident()] = weakref.ref(obj)

    def clear_threads(self,
                      existing_thread_ids,  # type: typing.Set[int]
                      ):
        """Clear the stored list of threads based on the list of existing thread ids.

        If any thread that is part of this list was stored, its data will be deleted.

        :param existing_thread_ids: A set of thread ids to keep.
        """
        # This code clears the thread/object mapping by clearing a copy and swapping it in an atomic operation This is
        # needed to be able to have this whole class lock-free and avoid concurrency issues.
        # The fact that it is lock free means we might lose some accuracy, but it's worth the trade-off for speed and simplicity.
        new_thread_id_to_object_mapping = self._thread_id_to_object.copy()
        # Iterate over a copy of the list of keys since it's mutated during our iteration.
        for thread_id in list(new_thread_id_to_object_mapping):
            if thread_id not in existing_thread_ids:
                del new_thread_id_to_object_mapping[thread_id]

        # Swap with the new list
        self._thread_id_to_object = new_thread_id_to_object_mapping

    def get_object(
            self,
            thread_id  # type: int
    ):
        # type: (...) -> typing.Optional[_T]
        """Return the object attached to thread.

        :param thread_id: The thread id.
        :return: The attached object.
        """
        obj_ref = self._thread_id_to_object.get(thread_id)
        if obj_ref is not None:
            return obj_ref()



# coding: utf-8

from msgpack import unpackb

from ddtrace.internal._encoding import packb


def test_str8():
    header = b'\xd9'
    data = b'x' * 32
    b = packb(data.decode())
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\x20'
    assert b[2:] == data
    assert unpackb(b, raw=True) == data

    data = b'x' * 255
    b = packb(data.decode())
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\xff'
    assert b[2:] == data
    assert unpackb(b, raw=True) == data


def test_bin8():
    header = b'\xc4'
    data = b''
    b = packb(data)
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\x00'
    assert b[2:] == data
    assert unpackb(b) == data

    data = b'x' * 255
    b = packb(data)
    assert len(b) == len(data) + 2
    assert b[0:2] == header + b'\xff'
    assert b[2:] == data
    assert unpackb(b) == data


def test_bin16():
    header = b'\xc5'
    data = b'x' * 256
    b = packb(data)
    assert len(b) == len(data) + 3
    assert b[0:1] == header
    assert b[1:3] == b'\x01\x00'
    assert b[3:] == data
    assert unpackb(b) == data

    data = b'x' * 65535
    b = packb(data)
    assert len(b) == len(data) + 3
    assert b[0:1] == header
    assert b[1:3] == b'\xff\xff'
    assert b[3:] == data
    assert unpackb(b) == data


def test_bin32():
    header = b'\xc6'
    data = b'x' * 65536
    b = packb(data)
    assert len(b) == len(data) + 5
    assert b[0:1] == header
    assert b[1:5] == b'\x00\x01\x00\x00'
    assert b[5:] == data
    assert unpackb(b) == data



"""
The pytest-bdd integration traces executions of scenarios and steps.

Enabling
~~~~~~~~

Please follow the instructions for enabling `pytest` integration.

.. note::
   The ddtrace.pytest_bdd plugin for pytest-bdd has the side effect of importing
   the ddtrace package and starting a global tracer.

   If this is causing issues for your pytest-bdd runs where traced execution of
   tests is not enabled, you can deactivate the plugin::

     [pytest]
     addopts = -p no:ddtrace.pytest_bdd

   See the `pytest documentation
   <https://docs.pytest.org/en/7.1.x/how-to/plugins.html#deactivating-unregistering-a-plugin-by-name>`_
   for more details.

"""

from ddtrace import config


# pytest-bdd default settings
config._add(
    "pytest_bdd",
    dict(
        _default_service="pytest_bdd",
    ),
)



"""
The molten web framework is automatically traced by ``ddtrace`` when calling ``patch``::

    from molten import App, Route
    from ddtrace import patch_all; patch_all(molten=True)

    def hello(name: str, age: int) -> str:
        return f'Hello {age} year old named {name}!'
    app = App(routes=[Route('/hello/{name}/{age}', hello)])


You may also enable molten tracing automatically via ``ddtrace-run``::

    ddtrace-run python app.py


Configuration
~~~~~~~~~~~~~

.. py:data:: ddtrace.config.molten['distributed_tracing']

   Whether to parse distributed tracing headers from requests received by your Molten app.

   Default: ``True``

.. py:data:: ddtrace.config.molten['service_name']

   The service name reported for your Molten app.

   Can also be configured via the ``DD_SERVICE`` or ``DD_MOLTEN_SERVICE`` environment variables.

   Default: ``'molten'``

:ref:`All HTTP tags <http-tagging>` are supported for this integration.

"""
from ...internal.utils.importlib import require_modules


required_modules = ["molten"]

with require_modules(required_modules) as missing_modules:
    if not missing_modules:
        from . import patch as _patch

        patch = _patch.patch
        unpatch = _patch.unpatch

        __all__ = ["patch", "unpatch"]



from concurrent import futures

from ddtrace.vendor.wrapt import wrap_function_wrapper as _w

from ..trace_utils import unwrap as _u
from .threading import _wrap_submit


def patch():
    """Enables Context Propagation between threads"""
    if getattr(futures, "__datadog_patch", False):
        return
    setattr(futures, "__datadog_patch", True)

    _w("concurrent.futures", "ThreadPoolExecutor.submit", _wrap_submit)


def unpatch():
    """Disables Context Propagation between threads"""
    if not getattr(futures, "__datadog_patch", False):
        return
    setattr(futures, "__datadog_patch", False)

    _u(futures.ThreadPoolExecutor, "submit")



from ddtrace import Pin
from ddtrace import config

from ...ext import SpanTypes
from ...internal.utils import get_argument_value
from ..trace_utils import unwrap
from ..trace_utils import with_traced_module
from ..trace_utils import wrap


config._add(
    "aiohttp_jinja2",
    dict(),
)


@with_traced_module
def traced_render_template(aiohttp_jinja2, pin, func, instance, args, kwargs):
    # original signature:
    # render_template(template_name, request, context, *, app_key=APP_KEY, encoding='utf-8')
    template_name = get_argument_value(args, kwargs, 0, "template_name")
    request = get_argument_value(args, kwargs, 1, "request")
    get_env_kwargs = {}
    if "app_key" in kwargs:
        get_env_kwargs["app_key"] = kwargs["app_key"]
    env = aiohttp_jinja2.get_env(request.app, **get_env_kwargs)

    # the prefix is available only on PackageLoader
    template_prefix = getattr(env.loader, "package_path", "")
    template_meta = "%s/%s" % (template_prefix, template_name)

    with pin.tracer.trace("aiohttp.template", span_type=SpanTypes.TEMPLATE) as span:
        span.set_tag("aiohttp.template", template_meta)
        return func(*args, **kwargs)


def _patch(aiohttp_jinja2):
    Pin().onto(aiohttp_jinja2)
    wrap("aiohttp_jinja2", "render_template", traced_render_template(aiohttp_jinja2))


def patch():
    import aiohttp_jinja2

    if getattr(aiohttp_jinja2, "_datadog_patch", False):
        return

    _patch(aiohttp_jinja2)

    setattr(aiohttp_jinja2, "_datadog_patch", True)


def _unpatch(aiohttp_jinja2):
    unwrap(aiohttp_jinja2, "render_template")


def unpatch():
    import aiohttp_jinja2

    if not getattr(aiohttp_jinja2, "_datadog_patch", False):
        return

    _unpatch(aiohttp_jinja2)

    setattr(aiohttp_jinja2, "_datadog_patch", False)



import os
from typing import List
from typing import Tuple

from .collector import ValueCollector
from .constants import CPU_PERCENT
from .constants import CPU_TIME_SYS
from .constants import CPU_TIME_USER
from .constants import CTX_SWITCH_INVOLUNTARY
from .constants import CTX_SWITCH_VOLUNTARY
from .constants import GC_COUNT_GEN0
from .constants import GC_COUNT_GEN1
from .constants import GC_COUNT_GEN2
from .constants import MEM_RSS
from .constants import THREAD_COUNT


class RuntimeMetricCollector(ValueCollector):
    value = []  # type: List[Tuple[str, str]]
    periodic = True


class GCRuntimeMetricCollector(RuntimeMetricCollector):
    """Collector for garbage collection generational counts

    More information at https://docs.python.org/3/library/gc.html
    """

    required_modules = ["gc"]

    def collect_fn(self, keys):
        gc = self.modules.get("gc")

        counts = gc.get_count()
        metrics = [
            (GC_COUNT_GEN0, counts[0]),
            (GC_COUNT_GEN1, counts[1]),
            (GC_COUNT_GEN2, counts[2]),
        ]

        return metrics


class PSUtilRuntimeMetricCollector(RuntimeMetricCollector):
    """Collector for psutil metrics.

    Performs batched operations via proc.oneshot() to optimize the calls.
    See https://psutil.readthedocs.io/en/latest/#psutil.Process.oneshot
    for more information.
    """

    required_modules = ["ddtrace.vendor.psutil"]
    stored_value = dict(
        CPU_TIME_SYS_TOTAL=0,
        CPU_TIME_USER_TOTAL=0,
        CTX_SWITCH_VOLUNTARY_TOTAL=0,
        CTX_SWITCH_INVOLUNTARY_TOTAL=0,
    )

    def _on_modules_load(self):
        self.proc = self.modules["ddtrace.vendor.psutil"].Process(os.getpid())

    def collect_fn(self, keys):
        with self.proc.oneshot():
            # only return time deltas
            # TODO[tahir]: better abstraction for metrics based on last value
            cpu_time_sys_total = self.proc.cpu_times().system
            cpu_time_user_total = self.proc.cpu_times().user
            cpu_time_sys = cpu_time_sys_total - self.stored_value["CPU_TIME_SYS_TOTAL"]
            cpu_time_user = cpu_time_user_total - self.stored_value["CPU_TIME_USER_TOTAL"]

            ctx_switch_voluntary_total = self.proc.num_ctx_switches().voluntary
            ctx_switch_involuntary_total = self.proc.num_ctx_switches().involuntary
            ctx_switch_voluntary = ctx_switch_voluntary_total - self.stored_value["CTX_SWITCH_VOLUNTARY_TOTAL"]
            ctx_switch_involuntary = ctx_switch_involuntary_total - self.stored_value["CTX_SWITCH_INVOLUNTARY_TOTAL"]

            self.stored_value = dict(
                CPU_TIME_SYS_TOTAL=cpu_time_sys_total,
                CPU_TIME_USER_TOTAL=cpu_time_user_total,
                CTX_SWITCH_VOLUNTARY_TOTAL=ctx_switch_voluntary_total,
                CTX_SWITCH_INVOLUNTARY_TOTAL=ctx_switch_involuntary_total,
            )

            metrics = [
                (THREAD_COUNT, self.proc.num_threads()),
                (MEM_RSS, self.proc.memory_info().rss),
                (CTX_SWITCH_VOLUNTARY, ctx_switch_voluntary),
                (CTX_SWITCH_INVOLUNTARY, ctx_switch_involuntary),
                (CPU_TIME_SYS, cpu_time_sys),
                (CPU_TIME_USER, cpu_time_user),
                (CPU_PERCENT, self.proc.cpu_percent()),
            ]

            return metrics



from gevent import monkey


monkey.patch_all()

import threading

from ddtrace.profiling import bootstrap
# do not use ddtrace-run; the monkey-patching would be done too late
import ddtrace.profiling.auto
from ddtrace.profiling.collector import stack_event


def fibonacci(n):
    if n == 0:
        return 0
    elif n == 1:
        return 1
    else:
        return fibonacci(n - 1) + fibonacci(n - 2)


# When not using our special PeriodicThread based on real threads, there's 0 event captured.
i = 1
while len(bootstrap.profiler._profiler._recorder.events[stack_event.StackSampleEvent]) < 10:
    threads = []
    for _ in range(10):
        t = threading.Thread(target=fibonacci, args=(i,))
        t.start()
        threads.append(t)
    i += 1
    for t in threads:
        t.join()



import os

import pyramid.config

from ddtrace import config
from ddtrace.vendor import wrapt

from ...internal.utils.formats import asbool
from .constants import SETTINGS_ANALYTICS_ENABLED
from .constants import SETTINGS_ANALYTICS_SAMPLE_RATE
from .constants import SETTINGS_DISTRIBUTED_TRACING
from .constants import SETTINGS_SERVICE
from .trace import DD_TWEEN_NAME
from .trace import trace_pyramid


config._add(
    "pyramid",
    dict(
        distributed_tracing=asbool(os.getenv("DD_PYRAMID_DISTRIBUTED_TRACING", default=True)),
    ),
)

DD_PATCH = "_datadog_patch"


def patch():
    """
    Patch pyramid.config.Configurator
    """
    if getattr(pyramid.config, DD_PATCH, False):
        return

    setattr(pyramid.config, DD_PATCH, True)
    _w = wrapt.wrap_function_wrapper
    _w("pyramid.config", "Configurator.__init__", traced_init)


def traced_init(wrapped, instance, args, kwargs):
    settings = kwargs.pop("settings", {})
    service = config._get_service(default="pyramid")
    # DEV: integration-specific analytics flag can be not set but still enabled
    # globally for web frameworks
    old_analytics_enabled = os.getenv("DD_PYRAMID_ANALYTICS_ENABLED")
    analytics_enabled = os.environ.get("DD_TRACE_PYRAMID_ANALYTICS_ENABLED", old_analytics_enabled)
    if analytics_enabled is not None:
        analytics_enabled = asbool(analytics_enabled)
    # TODO: why is analytics sample rate a string or a bool here?
    old_analytics_sample_rate = os.getenv("DD_PYRAMID_ANALYTICS_SAMPLE_RATE", default=True)
    analytics_sample_rate = os.environ.get("DD_TRACE_PYRAMID_ANALYTICS_SAMPLE_RATE", old_analytics_sample_rate)
    trace_settings = {
        SETTINGS_SERVICE: service,
        SETTINGS_DISTRIBUTED_TRACING: config.pyramid.distributed_tracing,
        SETTINGS_ANALYTICS_ENABLED: analytics_enabled,
        SETTINGS_ANALYTICS_SAMPLE_RATE: analytics_sample_rate,
    }
    # Update over top of the defaults
    # DEV: If we did `settings.update(trace_settings)` then we would only ever
    #      have the default values.
    trace_settings.update(settings)
    # If the tweens are explicitly set with 'pyramid.tweens', we need to
    # explicitly set our tween too since `add_tween` will be ignored.
    insert_tween_if_needed(trace_settings)

    # The original Configurator.__init__ looks up two levels to find the package
    # name if it is not provided. This has to be replicated here since this patched
    # call will occur at the same level in the call stack.
    if not kwargs.get("package", None):
        from pyramid.path import caller_package

        kwargs["package"] = caller_package(level=2)

    kwargs["settings"] = trace_settings
    wrapped(*args, **kwargs)
    trace_pyramid(instance)


def insert_tween_if_needed(settings):
    tweens = settings.get("pyramid.tweens")
    # If the list is empty, pyramid does not consider the tweens have been
    # set explicitly.
    # And if our tween is already there, nothing to do
    if not tweens or not tweens.strip() or DD_TWEEN_NAME in tweens:
        return
    # pyramid.tweens.EXCVIEW is the name of built-in exception view provided by
    # pyramid.  We need our tween to be before it, otherwise unhandled
    # exceptions will be caught before they reach our tween.
    idx = tweens.find(pyramid.tweens.EXCVIEW)
    if idx == -1:
        settings["pyramid.tweens"] = tweens + "\n" + DD_TWEEN_NAME
    else:
        settings["pyramid.tweens"] = tweens[:idx] + DD_TWEEN_NAME + "\n" + tweens[idx:]



import sys

from hypothesis import given
from hypothesis import strategies as st

from ddtrace.appsec._ddwaf import _Wrapper


SCALAR_OBJECTS = st.one_of(st.none(), st.booleans(), st.integers(), st.floats(), st.characters())

PYTHON_OBJECTS = st.recursive(
    base=SCALAR_OBJECTS,
    extend=lambda inner: st.lists(inner) | st.dictionaries(SCALAR_OBJECTS, inner),
)

WRAPPER_KWARGS = dict(
    max_objects=st.integers(min_value=0, max_value=2 ** 63 - 1),
)


@given(obj=PYTHON_OBJECTS, kwargs=st.fixed_dictionaries(WRAPPER_KWARGS))
def test_ddwaf_objects_wrapper(obj, kwargs):
    obj = _Wrapper(obj, **kwargs)
    repr(obj)
    del obj


if __name__ == "__main__":
    import atheris

    atheris.Setup(sys.argv, atheris.instrument_func(test_ddwaf_objects_wrapper.hypothesis.fuzz_one_input))
    atheris.Fuzz()



import os
import subprocess
import time

import pytest
import redis
import rq

from ddtrace import Pin
from ddtrace.contrib.rq import patch
from ddtrace.contrib.rq import unpatch
from tests.utils import override_config
from tests.utils import snapshot
from tests.utils import snapshot_context

from ..config import REDIS_CONFIG
from .jobs import JobClass
from .jobs import job_add1
from .jobs import job_fail


# Span data which isn't static to ignore in the snapshots.
snapshot_ignores = ["meta.job.id", "meta.error.stack"]

rq_version = tuple(int(x) for x in rq.__version__.split(".")[:3])


@pytest.fixture()
def connection():
    yield redis.Redis(port=REDIS_CONFIG["port"])


@pytest.fixture()
def queue(connection):
    patch()
    try:
        q = rq.Queue("q", connection=connection)
        yield q
    finally:
        unpatch()


@pytest.fixture()
def sync_queue(connection):
    patch()
    try:
        sync_q = rq.Queue("sync-q", is_async=False, connection=connection)
        yield sync_q
    finally:
        unpatch()


@snapshot(ignores=snapshot_ignores)
def test_sync_queue_enqueue(sync_queue):
    sync_queue.enqueue(job_add1, 1)


@snapshot(ignores=snapshot_ignores, variants={"": rq_version >= (1, 10, 1), "pre_1_10_1": rq_version < (1, 10, 1)})
def test_queue_failing_job(sync_queue):
    # Exception raising behavior was changed in 1.10.1
    # https://github.com/rq/rq/commit/93f34c796f541ea4b1c156426d6524df05753826
    if rq_version >= (1, 10, 1):
        sync_queue.enqueue(job_fail)
        return

    with pytest.raises(Exception):
        sync_queue.enqueue(job_fail)


@snapshot(ignores=snapshot_ignores)
def test_sync_worker(queue):
    job = queue.enqueue(job_add1, 1)
    worker = rq.SimpleWorker([queue], connection=queue.connection)
    worker.work(burst=True)
    assert job.result == 2


@snapshot(ignores=snapshot_ignores)
def test_sync_worker_multiple_jobs(queue):
    jobs = []
    for i in range(3):
        jobs.append(queue.enqueue(job_add1, i))
    worker = rq.SimpleWorker([queue], connection=queue.connection)
    worker.work(burst=True)
    assert [job.result for job in jobs] == [1, 2, 3]


@snapshot(ignores=snapshot_ignores)
def test_sync_worker_config_service(queue):
    job = queue.enqueue(job_add1, 10)
    with override_config("rq_worker", dict(service="my-worker-svc")):
        worker = rq.SimpleWorker([queue], connection=queue.connection)
        worker.work(burst=True)
    assert job.result == 11


@snapshot(ignores=snapshot_ignores)
def test_queue_pin_service(queue):
    Pin.override(queue, service="my-pin-svc")
    job = queue.enqueue(job_add1, 10)
    worker = rq.SimpleWorker([queue], connection=queue.connection)
    worker.work(burst=True)
    assert job.result == 11


@snapshot(ignores=snapshot_ignores)
def test_sync_worker_pin_service(queue):
    job = queue.enqueue(job_add1, 10)
    worker = rq.SimpleWorker([queue], connection=queue.connection)
    Pin.override(worker, service="my-pin-svc")
    worker.work(burst=True)
    assert job.result == 11


@snapshot(ignores=snapshot_ignores)
def test_worker_failing_job(queue):
    queue.enqueue(job_fail)
    worker = rq.SimpleWorker([queue], connection=queue.connection)
    worker.work(burst=True)


@snapshot(ignores=snapshot_ignores)
def test_worker_class_job(queue):
    queue.enqueue(JobClass().job_on_class, 2)
    queue.enqueue(JobClass(), 4)
    worker = rq.SimpleWorker([queue], connection=queue.connection)
    worker.work(burst=True)


@pytest.mark.parametrize("distributed_tracing_enabled", [False, None])
@pytest.mark.parametrize("worker_service_name", [None, "custom-worker-service"])
def test_enqueue(queue, distributed_tracing_enabled, worker_service_name):
    token = "tests.contrib.rq.test_rq.test_enqueue_distributed_tracing_enabled_%s_worker_service_%s" % (
        distributed_tracing_enabled,
        worker_service_name,
    )
    with snapshot_context(token, ignores=snapshot_ignores):
        env = os.environ.copy()
        env["DD_TRACE_REDIS_ENABLED"] = "false"
        if distributed_tracing_enabled is not None:
            env["DD_RQ_DISTRIBUTED_TRACING_ENABLED"] = str(distributed_tracing_enabled)
        if worker_service_name is not None:
            env["DD_SERVICE"] = "custom-worker-service"
        p = subprocess.Popen(["ddtrace-run", "rq", "worker", "q"], env=env)
        try:
            job = queue.enqueue(job_add1, 1)
            # Wait for job to complete
            for _ in range(100):
                if job.result is not None:
                    break
                time.sleep(0.1)
            assert job.result == 2
        finally:
            p.terminate()
            # Wait for trace to be sent
            time.sleep(0.5)
